; CALCULO DE PI Y LOGARITMO NATURAL DE 2 POR DARIO A. ALPERN
;
; Formulas utilizadas:
; ===================
;
; Calculo de pi:
;
;                   1            1            1             13753  
; pi = 4  6 arc tg  + arc tg  + arc tg  + arc tg  
;                   8           32           128          63664921 
;
;
; Calculo de ln 2:
;
;                    1             1             1529
; ln 2 =  5 arg tgh  + arg tgh  + arg tgh 
;                    8            16            670751
;
;
; El calculo del arco tangente se realiza mediante la formula de Taylor:
;
;                  3     5     7     9
;                 x     x     x     x
; arc tg x = x -  +  -  +  - ...
;                 3     5     7     9
;
; El calculo del argumento de la tangente hiperbolica se realiza mediante la
; formula de Taylor:
;
;                   3     5     7     9
;                  x     x     x     x
; arg tgh x = x +  +  +  +  + ...
;                  3     5     7     9
;
;
; Estas formulas se deducen teniendo en cuenta que:
;
;                                                          x + y
;   pi = 4 arc tg 1     y   arc tg x + arc tg y = arc tg  ;
;                                                         1 - xy
;
;                 3                                           x + y
; ln 2 = arg tgh     y   arg tgh x + arg tgh y = arg tgh  .
;                 5                                          1 + xy
;
; En el caso de procesadores de 16 bits es muy costoso en tiempo dividir por
; 63664921 o por 670751, por lo que se divide por sus factores (mediante dos
; divisiones). Las factorizaciones son: 63664921 = 1217 x 52313 y
; 670751 = 347 x 1933.
;
;----------------------------------------------------------------------------
;
; Durante el calculo de PI o LN 2, se utilizan tres buffers que contienen
; diferentes numeros en binario: el primero (que comienza en "buffer")
; contiene la suma parcial de los terminos; en el segundo (que comienza en
; "buffer + 16384") se calculan las potencias que intervienen en los
; diferentes terminos; mientras que en el ultimo (que comienza en
; "buffer + 32768") se calcula un termino que luego se sumara o restara del
; que comienza en "buffer".
;
;----------------------------------------------------------------------------
;
; Definicion de constantes y variables inicializadas del programa.
;
.386
LN2                equ 1
Formato            equ 2
MostrarCantDigitos equ 4
ASCII_CR           equ 13   ;Mover el cursor al extremo izquierdo.
ASCII_LF           equ 10   ;Mover el cursor hacia abajo.
port6845           equ 3D4h ;Port del controlador de video.

codigo  segment use16
        assume cs:codigo,ds:codigo
        org 100h            ;Offset de arranque del programa en formato .COM
comienzo:jmp Inicio         ;Salto al inicio del programa.

contador        db 0

divisor16       label word  ;Divisor de cada termino. Va tomando los valores
divisor32       dd 1        ;1, 3, 5,... como se aprecia en las formulas.

digitos16       label word  ;Cantidad de digitos que posee la aproximacion que
digitos32       dd 0        ;se desea mostrar.

XMS             label dword ;Direccion del manejador de XMS.
                dw 0
XMS_Instalado   dw 0        ;Si el segmento del manejador no es cero, el XMS
                            ;esta instalado.

CantWords       label word  ;Cantidad de palabras (16 bits en 8088, 80286 y
CantDWords      dd 0        ;32 bits en 80386) correspondientes a la cantidad
                            ;de digitos pedidos.

UltWord         label word
UltDWord        dd    0
PrimWord        label word  ;Palabra mas significativa distinta de cero.
PrimDWord       dd    0     ;Al no tener que procesar las palabras que valen
                            ;cero se duplica la velocidad.
Numerador       dw 13753    ;Numerador del argumento del ultimo arc tg. Esta
                            ;inicializado para el calculo de PI. En el caso
                            ;del calculo de ln 2 este valor se sobreescribe
                            ;con 1529 al principio del programa.

PrimerDenominador dw 1217   ;Primer factor del denominador del argumento del
                            ;ultimo arc tg. Esta inicializado para el calculo
                            ;de PI. En el caso del calculo de ln 2 este valor
                            ;se sobreescribe con 1933 al principio del
                            ;programa.

SegundoDenominador dw 52313 ;Segundo factor del denominador del argumento del
                            ;ultimo arc tg. Esta inicializado para el calculo
                            ;de PI. En el caso del calculo de ln 2 este valor
                            ;se sobreescribe con 347 al principio del
                            ;programa.

PunteroSubrPotenciasDe2 dw PotenciasBinarias16BitsParaPi
                            ;Puntero a la subrutina para calcular las potencias
                            ;binarias. Esta inicializada para el calculo de PI.
                            ;En el caso del calculo de ln 2 este valor se
                            ;sobreescribe con el correspondiente para ln 2.

handle          dw 1        ;Handle correspondiente al dispositivo estandar
                            ;de salida (la pantalla por defecto). Se
                            ;sobreescribe con el handle del archivo, si se
                            ;especifica.

Indicadores     db 0        ;Sirve para saber si la linea de comando contiene
                            ;los "switches" /pi o /ln2:
                            ;Bit 0 = 1: Calculo de ln 2.
                            ;Bit 0 = 0: Calculo de pi.
                            ;Bit 1 = 1: Salida formateada.
                            ;Bit 1 = 0: Salida sin formatear.

ContadorIzquierdo db 0

bufferdos       db "3,34567890123" ;Buffer para la conversion de binario a
                                   ;decimal.

BufferNumeroFormateado db "3,",75 dup (" "),ASCII_CR,ASCII_LF
                            ;Buffer de una linea para obtener la salida
                            ;formateada.

PosicionDigitoEnBuffer dw BufferNumeroFormateado + 2
                            ;Puntero dentro del buffer de una linea.

InformacionDeFormateado dw 0
DigitosMostrados32Bits  label dword
DigitosMostrados        dw 0,0
XMS_Handle              dw 0
LongitudBufferx16KB     dw 0

Iniciales     db "dhms"
Dias          db 0
Horas         db 0
Minutos       db 0
Segundos      db 0
FraccSegundo  dd 0

OldInt8Handler label dword
OldInt8Offset  dw 0
OldInt8Segment dw 0

MostrarEnInt8  db 1

CursorVideo    label dword
CursorVideoOffset  dw 0
CursorVideoSegment dw 0

;Estructura que necesita la funcion 0Bh de XMS (mover bytes).

EstructuraCopiarBytesXMS:
CantidadBytesAMover     dw 16384,0
HandleXMSFuente         dw 0          ;Cero si se apunta a RAM convencional.
OffsetXMSFuente         label dword
OffsetFuenteRAMConv     dw 0
SegmentoFuenteRAMConv   dw 0
HandleXMSDestino        dw 0          ;Cero si se apunta a RAM convencional.
OffsetXMSDestino        label dword
OffsetDestinoRAMConv    dw 0
SegmentoDestinoRAMConv  dw 0

CopyRight db "Calculo de PI y de ln 2. Hecho por Dario Alejandro Alpern."
        db ASCII_CR,ASCII_LF
LongitudCopyRight equ $ - CopyRight
Texto   db ASCII_CR,ASCII_LF
        db "Uso del programa: PI cantidad_de_decimales [/F] [/LN2]"
        db " [archivo_destino]",ASCII_CR,ASCII_LF
        db "   /F: Muestra la salida formateada para editores de texto"
        db ASCII_CR,ASCII_LF
        db "   /LN2: Muestra el valor del logaritmo natural de 2 en vez de pi"
        db ASCII_CR,ASCII_LF
        db "   archivo_destino: Archivo en que se almacenara el numero."
        db ASCII_CR,ASCII_LF
        db "                    Si no se especifica, es la pantalla (STDOUT)."
        db ASCII_CR,ASCII_LF
        db ASCII_CR,ASCII_LF
        db "En el caso de usar 8088, 80286 o no estar instalado un manejador"
        db " de XMS",ASCII_CR,ASCII_LF
        db "(DEVICE=C:\DOS\HIMEM.SYS o similar en el archivo CONFIG.SYS) "
        db "la cantidad",ASCII_CR,ASCII_LF
        db "maxima de decimales es 39400.",ASCII_CR,ASCII_LF
LongitudTexto equ $ - Texto
MemoriaLlena db ASCII_CR,ASCII_LF
        db "No hay suficiente memoria extendida para procesar tantos "
        db "decimales.",ASCII_CR,ASCII_LF
LongitudMemoriaLlena equ $ - MemoriaLlena
TextoSegundaLinea db "Digitos hallados: "
LongitudTextoSegundaLinea equ $ - TextoSegundaLinea
TiempoTotal db ASCII_CR,ASCII_LF,"Tiempo total: "
LongitudTiempoTotal equ $ - TiempoTotal
;=============================================================================
; Determinacion del procesador.
;=============================================================================
Inicio: mov dx,offset CopyRight ;Mostrar en pantalla el "copyright".
        mov cx,LongitudCopyRight
        mov bx,2                ;Handle STDERR.
        mov ah,40h
        int 21h
test86: pushf           ;Sirve para saber si la CPU es un 8086 u 8088. En este
        pop ax          ;procesador los bits 15-12 del registro de indicadores
        and ax,0FFFh    ;siempre estan a uno.
        push ax
        popf
        pushf
        pop ax
        add ax,1000h
        jc short Inicio16
test286:pushf           ;Sirve para saber si la CPU es un 80286. En este
        pop ax          ;procesador los bits 15-12 del registro de indicadores
        or ax,0F000h    ;siempre estan a cero en modo real.
        push ax
        popf
        pushf
        pop ax
        and ax,0F000h
        jz short Inicio16
        jmp Inicio32    ;El procesador es un 80386 o posterior.
;=============================================================================
; Calculo de PI y LN 2 con procesadores de 16 bits.
;=============================================================================
;
; Obtencion de la cantidad de digitos a calcular.
;
Inicio16:
        mov bx,80h              ;Posicion anterior al inicio de la linea de
                                ;comando.
        mov si,10               ;Multiplicador para convertir decimal a binario
        mov ah,0                ;El byte mas significativo de un digito siempre
                                ;es cero.
        mov ch,0                ;Lo mismo.
BusquedaDeDigito16:
        inc bx                  ;Apuntar al siguiente byte.
        mov al,[bx]             ;Obtener dicho byte.
        cmp al,ASCII_CR         ;Fin de la linea?
        je short error          ;Error si es asi (no hay numero).
        cmp al," "              ;Es espacio o "tab"?
        jbe BusquedaDeDigito16  ;Saltar si es asi.
        sub al,"0"              ;Convertir el digito a binario.
        cmp al,9                ;Es un digito?
        jbe short NumeroLineaDeComando16 ;Saltar si lo es.
error:  mov dx,offset Texto     ;Mostrar texto indicando el formato de la
        mov cx,LongitudTexto    ;linea de comando.
        mov bx,2                ;Handle STDERR.
        mov ah,40h
        int 21h
        mov ax,4c01h            ;Fin del programa con codigo de error 1.
        int 21h
NuevoDigito16:
        cmp ax,3940             ;Al multiplicar por 10 se superara el maximo?
        jae short error         ;Error si es asi.
        mul si                  ;Multiplicar el numero por 10.
        add ax,cx               ;Sumar el digito.
NumeroLineaDeComando16:
        inc bx                  ;Apuntar al siguiente byte de la linea de com.
        mov cl,[bx]             ;Obtener dicho byte.
        sub cl,"0"              ;Convertir el digito a binario.
        cmp cl,9                ;Es un digito?
        jbe NuevoDigito16       ;Saltar si es asi.
        mov digitos16,ax        ;Almacenar la cantidad de digitos en la
                                ;variable correspondiente.
        mov si,13607            ;2^16 / log   (2^16)
        mul si                  ;          10
                                ;DX tiene la cantidad de palabras correspon-
                                ;dientes a esa cantidad de digitos.
        add dx,5                ;Agregar algunas palabras para tener en cuenta
                                ;el truncamiento.
        mov CantWords,dx        ;Almacenar la cantidad de palabras.
;
;------------------------------------------------------------------------------
;
; Ver si en la linea de comandos aparecen /F o /LN2. En el caso que existan,
; se ponen a uno ciertos bits de la variable Indicadores. Para /F se pone a
; 1 el bit Formato, mientras que para /LN2 se pone a 1 el bit LN2.
;
        call switch
;
;------------------------------------------------------------------------------
;
; Determinacion de constantes para calculo de PI o de LN 2. En el caso de PI
; se utilizan las constantes definidas mediante DB o DW, mientras que para el
; otro caso se cargan dichas constantes en las variables indicadas como se
; muestra a continuacion.
;
        mov si,4*13753/1217     ;Entero de 1ra division del 1er termino de PI.
        mov dx,4*13753 mod 1217 ;Palabra mas significativa de la fraccion.
        test Indicadores,LN2    ;Ver si se esta calculando PI o LN 2.
        jz short FinInicializacionCtes  ;Saltar si se esta calculando PI.
CalculoDeLN2:
        xor si,si             ;Entero de 1ra division del 1er termino de LN 2.
        mov dx,1529             ;Palabra mas significativa de la fraccion.
        mov Numerador,dx        ;Numerador para LN 2.
        mov PrimerDenominador,1933 ;Primer denominador para LN 2.
        mov SegundoDenominador,347 ;Segundo denominador para LN 2.
        mov PunteroSubrPotenciasDe2,offset PotenciasBinarias16BitsParaLN2
                                ;Subrutina a llamar cuando se calculen las
                                ;potencias de 2.
        mov bufferdos,"0"       ;Indicar que el numero comienza con cero.
        mov BufferNumeroFormateado,"0" ;Lo mismo.
FinInicializacionCtes:
;
;----------------------------------------------------------------------------
;
; Calculo del primer termino. Al finalizar los buffers que comienzan en
; "buffer" y en "buffer+16384" contendran el primer termino en binario (solo
; la parte fraccionaria ya que la entera no se incluye en el caso de PI).
;
; En esta seccion se calcularan ambas divisiones multipalabra en un solo
; ciclo.
; Cuando empieza esta parte del programa, DX contiene el resto de dividir el
; numerador por el primer factor del denominador.
        mov cx,CantWords          ;Cantidad de palabras a calcular.
        mov di,offset buffer      ;Puntero al inicio del buffer.
        mov bx,PrimerDenominador  ;Obtener los factores que componen el
        mov bp,SegundoDenominador ;denominador.
CicloCalculoPrimerTermino:
        xor ax,ax  ;Como el numerador es un entero, solo hay ceros en la parte
                   ;fraccionaria, por lo que en el dividendo solo se agregan
                   ;ceros.
        div bx     ;Efectuar la division por el primer factor del denominador.
                   ;En este momento DX contiene el resto y AX el cociente, por
                   ;lo que este ultimo deberia ir al buffer. Esto no es nece-
                   ;sario pues es justamente el byte que se utilizaria en la
                   ;division con el segundo factor del denominador.
        xchg dx,si ;Intercambiar el resto de la primera division con el de la
                   ;segunda.
        div bp     ;Efectuar la division por el segundo factor del denominador.
                   ;Ahora si debera guardarse AX (el cociente) en el buffer.
        xchg dx,si ;Intercambiar el resto de la primera division con el de la
                   ;segunda.
        mov [di+16384],ax ;Guardar la palabra del cociente en el buffer que
                   ;comienza en "buffer+16384".
        stosw      ;Guardar la palabra del cociente en el buffer que comienza
                   ;en "buffer". Ademas incrementa el puntero.
        loop CicloCalculoPrimerTermino ;Fin de la division.
;
;-----------------------------------------------------------------------------
;
        test Indicadores,LN2              ;Se debe calcular ln 2 o PI?
        jnz short FinalPrimerTermino16BitsLN2   ;Saltar si se debe calcular
                                                ;ln 2.
;
; Sumar las potencias de 2 correspondientes al primer termino tanto en el
; resultado (que comienza en "buffer") como al buffer que contiene el termino
; que se esta calculando (en "buffer+16384"). En este caso, como se desprende
; de la formula que figura mas arriba, se debe sumar 4/32+4/128 a 4*13753/
; 63664921 (ya calculado previamente). Como 6*4/8 = 3, la parte fraccionaria
; es cero y no se considera. Como este ultimo termino es menor que 1/256,
; puede utilizarse la instruccion MOV para cargar el byte mas significativo
; (si el termino hubiera sido >= 1/256, deberia haberse utilizado la
; instruccion OR).
;
        mov byte ptr buffer[1],256*4/32+256*4/128
        mov byte ptr buffer[16384+1],256*4/32+256*4/128
;
;-----------------------------------------------------------------------------
;
; Ciclo principal del calculo de PI. Cada ciclo incluye primero una resta
; (ver terminos pares de la formula) y luego una suma (terminos impares).
;
CicloPrincipal16BitsPi:
;
; Resta de terminos pares al resultado.
;
        call NuevoTermino16Bits   ;En el buffer que comienza en "buffer +
                                  ;32768" se obtiene el termino a restar.
CicloRestaDelTermino16BitsPi:     ;Ciclo que realiza la resta multipalabra.
        mov ax,[di]               ;Obtener palabra del minuendo (de "buffer").
        sbb ax,[si]               ;Restar sustraendo con carry (de "buffer +
                                  ;32768").
        stosw                     ;Poner el resultado donde estaba el minuendo.
        dec si                    ;Ajustar el puntero.
        dec si
        loop CicloRestaDelTermino16BitsPi ;Cerrar el ciclo.
;
; En este momento faltan restar las palabras mas significativas del resultado
; con el "carry" pues el sustraendo es cero.
;
CicloSeguirRestando16BitsPi:
        sbb [di],cx               ;Sustraer cero con prestamo.
        dec di                    ;Ajustar el puntero sin alterar el CF.
        dec di
        jc CicloSeguirRestando16BitsPi ;Volver atras si todavia hay carry.
;
; Suma de terminos impares al resultado.
;
        call NuevoTermino16Bits   ;En el buffer que comienza en "buffer +
                                  ;32768" se obtiene el termino a sumar.
CicloSumaDelTermino16BitsPi:      ;Ciclo que realiza la suma multipalabra.
        lodsw                     ;Obtener palabra del termino a sumar.
        adc ax,[di]               ;Sumarselo al resultado.
        stosw                     ;Poner nueva palabra en el resultado.
        loop CicloSumaDelTermino16BitsPi ;Cerrar el ciclo.
;
; En este momento faltan sumar las palabras mas significativas del resultado
; con el "carry" pues el sustraendo es cero.
;
CicloSeguirSumando16BitsPi:
        adc [di],cx               ;Sumar cero con "carry"
        dec di                    ;Ajustar puntero.
        dec di
        jc CicloSeguirSumando16BitsPi ;Volver atras si todavia hay carry.
        jmp CicloPrincipal16BitsPi ;Ir a procesar el siguiente termino.
;
;----------------------------------------------------------------------------
;
; Sumar las potencias de 2 correspondientes al primer termino tanto en el
; resultado (que comienza en "buffer") como al buffer que contiene el termino
; que se esta calculando (en "buffer+16384"). En este caso, como se desprende
; de la formula que figura mas arriba, se debe sumar 5/8+1/16 a 1529/670751.
; (ya calculado previamente). Como este ultimo termino es menor que 1/256,
; puede utilizarse la instruccion MOV para cargar el byte mas significativo
; (si el termino hubiera sido >= 1/256, deberia haberse utilizado la
; instruccion OR).
;
FinalPrimerTermino16BitsLN2:
        mov byte ptr buffer[1],256*5/8+256/16
        mov byte ptr buffer[16384+1],256*5/8+256/16
;
;----------------------------------------------------------------------------
;
; Ciclo principal del calculo de ln 2. Cada ciclo incluye una suma con cada
; termino (ver formula).
;
CicloPrincipal16BitsLN2:
        call NuevoTermino16Bits   ;En el buffer que comienza en "buffer +
                                  ;32768" se obtiene el termino a sumar.
CicloSumaDelTermino16BitsLN2:     ;Ciclo que realiza la suma multipalabra.
        lodsw                     ;Obtener palabra del termino a sumar.
        adc ax,[di]               ;Sumarselo al resultado.
        stosw                     ;Poner nueva palabra en el resultado.
        loop CicloSumaDelTermino16BitsLN2 ;Cerrar el ciclo.
;
; En este momento faltan sumar las palabras mas significativas del resultado
; con el "carry" pues el sustraendo es cero.
;
CicloSeguirSumando16BitsLN2:
        adc [di],cx               ;Sumar cero con "carry"
        dec di                    ;Ajustar puntero.
        dec di
        jc CicloSeguirSumando16BitsLN2 ;Volver atras si todavia hay carry.
        jmp CicloPrincipal16BitsLN2 ;Ir a procesar el siguiente termino.
;
;----------------------------------------------------------------------------
;
; Calculo de las potencias de 2 (segun la formula, 5/(8^N)+1/(16^N) donde N
; es el divisor. Como se utiliza la instruccion XOR, cambiando bits, esta
; subrutina sirve para poner y para sacar estas potencias de 2.
;
PotenciasBinarias16BitsParaLN2:
;
; Calculo de 4/(8^N). Para ello debera cambiarse el bit 2-3N siendo el cero
; el menos significativo a la izquierda del punto decimal. El signo negativo
; del bit indica que corresponde a la derecha del punto decimal.
;
        mov ax,08000h             ;Indicar bit -1.
        mov di,divisor16          ;Obtener el valor de N.
        dec di                    ;Restarle uno.
        mov cx,di
        mov bh,al                 ;BH:CX = N-1
        add di,cx
        adc bh,al                 ;BH:DI = 2N-2
        add di,cx
        adc bh,al                 ;BH:DI = 3N-3
        mov cx,di                 ;BH:CX = 3N-3
        ror ax,cl                 ;Correr hacia la derecha 3N-3 lugares,
                                  ;obteniendo el bit -1-(3N-3) = 2-3N.
;
; Obtener la posicion de memoria donde debe aplicarse la mascara. Para ello
; se debe dividir el numero de bit por 8 y luego poner el bit menos
; significativo a cero (porque se trabaja con palabras).
;
        shr bh,1
        rcr di,1
        shr di,1
        shr di,1                  ;DI = BH:DI / 8.
        and di,0FFFEh             ;Poner el bit menos significativo a cero.
        add di,offset buffer[16384] ;Sumar el inicio de la zona a cambiar.
        xor [di],ax               ;Cambiar el bit 2-3N.
        mov ax,3                  ;Cambiar el bit -3N (correspondiente a
        call Potencia16Bits       ;1/(8^N)).
        mov ax,4                  ;Cambiar el bit -4N (correspondiente a
        jmp short Potencia16Bits  ;1/(16^N)).
;
;---------------------------------------------------------------------------
;
; Calculo de las potencias de 2 (segun la formula, 24/(8^N)+4/(32^N)+4/(128^N)
; donde N es el divisor. Como se utiliza la instruccion XOR, cambiando bits,
; esta subrutina sirve para poner y para sacar estas potencias de 2.
;
PotenciasBinarias16BitsParaPi:
;
; Calculo de 24/(8^N). Para ello debera cambiarse los bits 4-3N y 3-3N siendo
; el cero el menos significativo a la izquierda del punto decimal. El signo
; negativo del bit indica que corresponde a la derecha del punto decimal.
;
        mov ax,0C000h               ;Indicar bits -1 y -2.
        mov di,divisor16            ;Obtener el valor de N.
        mov cx,di
        mov bh,al                   ;BH:CX = N.
        add di,cx
        adc bh,al                   ;BH:DI = 2N.
        add di,cx
        adc bh,al                   ;BH:DI = 3N.
        sub di,5
        sbb bh,al                   ;BH:DI = 3N-5.
        jb short OtrasPotencias16BitsParaPi ;Saltar si 3N-5 es menor que cero.
                                    ;Esto ocurre solo en el primer termino,
                                    ;para el cual 24/8 = 3 no tiene parte
                                    ;fraccionaria, por lo que no hay que
                                    ;procesarlo.
        mov cx,di                   ;BH:CX = 3N-5.
        ror ax,cl                   ;Correr hacia la derecha 3N-5 lugares,
                                    ;obteniendo los bit -1-(3N-5) = 4-3N y
                                    ;-2-(3N-5) = 3-3N.
;
; Obtener la posicion de memoria donde debe aplicarse la mascara. Para ello
; se debe dividir el numero de bit por 8 y luego poner el bit menos
; significativo a cero (porque se trabaja con palabras).
;
        shr bh,1
        rcr di,1
        shr di,1
        shr di,1                    ;DI = BH:DI / 8.
        and di,0FFFEh               ;Poner el bit menos significativo a cero.
        add di,offset buffer[16384] ;Sumar el inicio de la zona a cambiar.
        xor [di],ax                 ;Cambiar los bits 4-3N y 3-3N.
OtrasPotencias16bitsParaPi:
        mov ax,5                    ;Cambiar el bit 2-5N (correspondiente a
        call Potencia16Bits         ;4/(32^N)).
        mov ax,7                    ;Cambiar el bit 2-7N (correspondiente a
                                    ;4/(128^N)).
;----------------------------------------------------------------------------
;
; Esta subrutina cambia el bit a - b * N, donde a=0 si se esta calculando ln 2
; y a=2 si se calcula PI, b esta representado por el registro AX, y N es el
; divisor como aparece en las formulas de Taylor dadas mas arriba.
;
Potencia16Bits:
        mul divisor16               ;DX:AX = b*N.
        test Indicadores,LN2        ;Se esta calculando ln 2 o PI?
        jnz short NoRestar16Bits    ;Saltar si se esta calculando ln 2.
        sub ax,2
        sbb dx,0                    ;DX:AX = b*N-2
NoRestar16Bits:
        mov bx,16
        div bx                      ;AX = Numero de palabra que contiene el
                                    ;bit a cambiar.
        mov cl,dl                   ;CL = Numero de bit dentro de la palabra.
        mov dx,1
        ror dx,cl                   ;DX = Mascara correspondiente al bit.
        cmp ax,CantWords            ;La palabra que contiene el bit a cambiar
                                    ;esta dentro del numero?
        jae short FinSubrutina16    ;Saltar si no es asi (la fraccion es
                                    ;insignificante).
        add ax,ax                   ;Convertir numero de palabra a numero de
                                    ;byte del bit a cambiar (con el bit menos
                                    ;significativo a cero porque estamos
                                    ;trabajando con palabras).
        mov si,ax                   ;Poner el indice en un registro que se
                                    ;pueda usar para acceso indirecto a
                                    ;memoria.
        xor word ptr buffer[si+16384],dx ;Cambiar el bit.
FinSubrutina16:
        ret                         ;Fin de la subrutina.
;
;---------------------------------------------------------------------------
;
; Calculo del nuevo termino (sin realizar la division por 3,5,7,...).
; El termino anterior se encuentra a partir de "buffer+16384" al igual que
; el termino a calcular. Esta subrutina realiza las siguientes tareas:
; 1) Retirar las potencias de 2 del termino anterior.
; 2) Multiplicar por el cuadrado de 13753/63664921 o 1529/670751 segun que
;    se este calculando PI o ln 2, respectivamente.
; 3) Agregar las potencias de 2 del nuevo termino.
NuevoTermino16Bits:
        call PunteroSubrPotenciasDe2  ;Retirar potencias de 2 del termino
                                    ;anterior.
        mov bp,2                    ;Cantidad de veces que debe realizarse la
                                    ;division en el paso 2.
;
; Ciclo para realizar la division.
;
CicloCalculoCociente:
        mov di,offset buffer[16384] ;Apuntar al inicio del buffer.
        mov ax,PrimWord             ;AX = Palabra mas significativa distinta
                                    ;de ceros.
        mov cx,CantWords            ;CX = Cantidad de palabras del numero.
        sub cx,ax                   ;CX = Cantidad de palabras a procesar.
        jbe short CantidadPalabrasDelTerminoOk16 ;Saltar si es <= 0.
        add di,ax
        add di,ax                   ;DI = Puntero a la primera palabra a la
                                    ;que debe realizarse la division.
        push bp                     ;Preservar contador.
        cld                         ;Procesar el buffer hacia posiciones
                                    ;superiores de memoria.
        xor si,si                   ;Resto de la division por el 2do factor.
        xor dx,dx                   ;Resto de la division por el 1er factor.
        mov bx,PrimerDenominador    ;BX = Primer factor del denominador.
        mov bp,SegundoDenominador   ;BP = Segundo factor del denominador.
        push cx                     ;Preservar cantidad de palabras a procesar.
CicloDiv16BitsPorDenominador:
        mov ax,[di]                 ;Obtener palabra del dividendo.
        div bx                      ;Dividir por 1er factor del denominador.
        xchg dx,si                  ;Intercambiar restos.
        div bp                      ;Dividir por 2do factor del denominador.
        xchg dx,si                  ;Intercambiar restos.
        stosw                       ;Guardar palabra del divisor.
        loop CicloDiv16BitsPorDenominador ;Cerrar ciclo de division
                                    ;multipalabra.
;
; Multiplicacion del cociente anterior por el numerador.
;
        mov si,Numerador            ;Obtener numerador.
        pop cx                      ;Restaurar cantidad de palabras a procesar.
        std                         ;Procesar el buffer hacia posiciones
                                    ;inferiores de memoria.
        dec di
        dec di                      ;DI = Puntero a la ultima palabra del
                                    ;cociente.
        xor bx,bx                   ;BX = Arrastre de la multiplicacion
                                    ;anterior.
        push cx                     ;Preservar cantidad de palabras a procesar.
CicloMult16BitsPorNumerador:
        mov ax,[di]                 ;Obtener palabra del cociente.
        mul si                      ;Multiplicar por factor de la formula de
                                    ;Taylor.
        add ax,bx                   ;Agregar arrastre de la multiplicacion
                                    ;anterior.
        adc dx,0                    ;Ajustar byte mas significativo del
                                    ;producto.
        mov bx,dx                   ;Arrastre de esta multiplicacion.
        stosw                       ;Poner la palabra en el buffer.
        loop CicloMult16BitsPorNumerador ;Cerrar ciclo de multiplicacion
                                    ;multipalabra.
        pop cx                      ;Restaurar cantidad de palabras a procesar.
        inc di
        inc di                      ;DI = Puntero a la palabra mas
                                    ;significativa distinta de ceros.
        pop bp                      ;Restaurar contador.
        and ax,ax                   ;La palabra es cero?
        jnz short FInCicloCalculoCociente ;Saltar si no es asi.
        inc PrimWord                ;La palabra mas significativa distinta de
                                    ;cero es la siguiente.
FinCicloCalculoCociente:
        dec bp                      ;Decrementar contador.
        jnz CicloCalculoCociente    ;Cerrar bucle del paso 2.
CantidadPalabrasDelTerminoOk16:
        add divisor16,2             ;Sumarle 2 al divisor (3,5,7...) en la
                                    ;formula de Taylor.
        call PunteroSubrPotenciasDe2 ;Agregar las potencias de 2 de este
                                    ;termino.
;
; Verificar si se termino el calculo en binario y hay que comenzar a realizar
; la conversion a decimal.
;
        mov cx,CantWords            ;Cantidad de palabras del numero.
        lea si,[di+offset comienzo - 100h -(offset buffer[16384])]
                                    ;SI = Cantidad de bytes a cero del numero.
        shr si,1                    ;SI = Cantidad de palabras a cero.
        sub cx,si                   ;CX = Cantidad de palabras significativas.
        jbe short BinarioADecimal16 ;Saltar si todas las palabras son cero.
                                    ;Esto quiere decir que el termino vale cero
                                    ;y que por lo tanto se acaba la serie de
                                    ;Taylor dentro de la aproximacion pedida.
;
; Dividir el termino por el divisor de la formula de Taylor (3, 5, 7, ...)
;
        xor dx,dx                   ;Poner el MSW del dividendo a cero.
        mov bx,divisor16            ;Obtener el divisor.
        cld                         ;Procesar los buffers dividendo y divisor
                                    ;hacia direcciones crecientes.
        mov si,di                   ;SI = Puntero a la primera palabra
                                    ;significativa.
        add di,16384                ;Apuntar al buffer del cociente.
        push cx                     ;Preservar la cantidad de digitos
                                    ;significativos del dividendo.
CicloDivisionPorValorVariable16:
        lodsw                       ;AX = Palabra del dividendo.
        div bx                      ;Realizar la division de 16 bits.
                                    ;AX <- DX:AX / BX.
                                    ;En DX aparece el resto de la division.
                                    ;En el proximo ciclo sera la palabra mas
                                    ;significativa del dividendo.
        stosw                       ;Almacenar la palabra del divisor.
        loop CicloDivisionPorValorVariable16 ;Cerrar el ciclo.
        pop cx                      ;Restaurar la cantidad de digitos
                                    ;significativos del dividendo.
        mov di,CantWords            ;DI = Cantidad de palabras del numero.
        add di,di                   ;DI = Cantidad de bytes del numero.
        lea si,buffer[di+32768-2]   ;SI = Puntero a la palabra menos
                                    ;significativa del termino.
        add di,offset buffer[-2]    ;SI = Puntero a la palabra menos
                                    ;significativa de la suma.
        std                         ;Procesar los buffers dividendo y divisor
                                    ;hacia direcciones decrecientes.
        ret                         ;Final de la subrutina.
;
;---------------------------------------------------------------------------
; Conversion de binario a decimal enviando el resultado al dispositivo
; estandar de salida. La conversion se realiza de a seis digitos por vez,
; ya que se multiplica el numero por 62 500 = 1 000 000 / 16.
;
BinarioADecimal16:
        pop ax                      ;Eliminar la direccion de retorno.
        test Indicadores,Formato    ;La salida debe estar formateada?
        jnz short NoMostrarEntero   ;Saltar si es asi.
        mov dx,offset bufferdos     ;DX = Puntero al numero y la coma.
        mov cx,2                    ;CX = Cantidad de bytes a mostrar.
        mov bx,handle               ;BX = Handle del dispositivo de salida.
        mov ah,40h                  ;AH = Funcion de escritura.
        int 21h                     ;Enviar el digito y la coma al dispositivo.
NoMostrarEntero:
        std                         ;Procesar el buffer en direcciones
                                    ;decrecientes.
        mov ax,CantWords            ;AX = Cantidad de palabras del numero.
        add ax,ax                   ;AX = Cantidad de bytes del numero.
        add ax,offset buffer[-2]    ;AX = Puntero a la palabra menos
                                    ;significativa del numero.
        mov UltWord,ax              ;Indicar donde se encuentra la palabra
                                    ;menos significativa del numero.
        call MultiplicarPor62500    ;Multiplicar el numero por 62500.
CicloBinarioADecimal16:
        mov ax,[di+2]
        and byte ptr [di+3],0Fh
        add ax,ax
        adc dx,dx
        adc ax,ax
        adc dx,dx
        adc ax,ax
        adc dx,dx
        adc ax,ax
        adc dx,dx
        adc ax,ax
        and al,0Fh
        call Convertir6Digitos
        mov al,dh
        mov dh,dl
        mov dl,[di+3]
        mov byte ptr [di+3],0
        call Convertir6Digitos
        mov ax,[di+2]
        and word ptr [di+2],0Fh
        add ax,ax
        adc dx,dx
        add ax,ax
        adc dx,dx
        add ax,ax
        adc dx,dx
        add ax,ax
        adc dx,dx
        mov al,dh
        mov dh,dl
        mov dl,ah
        call Convertir6Digitos
        mov al,dl
        mov dx,[di+2]
        dec CantWords
        call Convertir6Digitos
        jmp CicloBinarioADecimal16
Convertir6Digitos:                  ;Los digitos a convertir estan en AL:DX.
        test Indicadores,Formato
        jnz short Convertir6DigitosEnDI
        mov di,offset bufferdos
Convertir6DigitosEnDI:
        mov ah,0
        mov si,34464
        mov bl,1
        call ConvertirUnDigito16Bits
        mov si,10000
        dec bx
        call ConvertirUnDigito16Bits
        mov si,1000
        call ConvertirUnDigito16Bits
        mov si,100
        call ConvertirUnDigito16Bits
        mov si,10
        call ConvertirUnDigito16Bits
        mov si,1
        call ConvertirUnDigito16Bits
        test Indicadores,MostrarCantDigitos
        jnz short FinalConvertir6Digitos
        test Indicadores,Formato
        jnz short MultiplicarPor62500
        mov cx,digitos16
        cmp cx,6
        jbe short CantidadDeDigitosAMostrarOK16
        mov cx,6
CantidadDeDigitosAMostrarOK16:
        mov bx,handle
        mov dx,offset bufferdos
        mov ah,40h
        int 21h
        sub digitos16,6
        ja short MultiplicarPor62500
        mov ax,4c00h
        int 21h
;---------------------------------------------------------------------------
; Multiplicacion del numero por 1 000 000 / 16 = 62 500.
;
MultiplicarPor62500:
        mov cx,CantWords            ;CX = Cantidad de palabras del numero.
        mov di,UltWord              ;DI = Puntero a la palabra menos signific.
        mov si,62500                ;SI = Multiplicador.
        xor bx,bx                   ;BX = Arrastre desde la palabra anterior.
                                    ;Como no hay palabra anterior, vale cero.
CicloMultPor62500:
        mov ax,[di]                 ;AX = Palabra del numero.
        mul si                      ;DX:AX <- AX * SI.
        add ax,bx                   ;Sumar el arrastre desde la palabra
                                    ;anterior.
        adc dx,0                    ;Ajustar la palabra mas significativa.
        mov bx,dx                   ;Esta palabra sera el arrastre hacia la
                                    ;palabra siguiente.
        stosw                       ;Almacenar la palabra menos significativa
                                    ;del producto.
        loop CicloMultPor62500      ;Cerrar el ciclo.
        dec contador
        jns short DecrementarCantidadDePalabras16
        mov contador,15
FinalConvertir6Digitos:
        ret                         ;Final de la subrutina.
DecrementarCantidadDePalabras16:
        sub UltWord,2
        dec CantWords
        ret                         ;Final de la subrutina.
;-------------------------------------------------------------------------
; Subrutina ConvertirUnDigito16Bits: Halla un digito del numero y, si esta
; habilitado con el switch /F, realiza el formateo de texto.
;
; Entrada: BL:SI = Orden del digito a convertir (100000, 10000, 1000, 100,
;                  10, 1).
;          DI = Puntero al buffer donde se almacenan los digitos en ASCII.
;          AL:DX = Numero a convertir a decimal.
;          AH = 0: Mostrar los ceros mas significativos como espacios.
;          AH = 1: Mostrar los ceros mas significativos como ceros.
ConvertirUnDigito16Bits:
        mov bh,"0"-1
CicloConvertirUnDigito16Bits:
        inc bh
        sub dx,si
        sbb al,bl
        jnc CicloConvertirUnDigito16Bits
        add dx,si
        adc al,bl                     ;BH = Digito en ASCII.
                                      ;AL:DX = Nuevo numero a convertir.
        test Indicadores,MostrarCantDigitos
        jnz short MostrarUnDigito16Bits
        test Indicadores,Formato
        jnz short MostrarDigitoFormateado16
HuboDigitosSignificativos16Bits:
        mov ah,1                      ;Salida sin formatear.
PonerElDigitoEnBuffer16Bits:
        mov [di],bh                   ;Copiar el digito al buffer de salida.
        inc di                        ;Incrementar puntero buffer de salida.
        ret
MostrarUnDigito16Bits:
        cmp bh,"0"                    ;El digito es cero?
        jnz HuboDigitosSignificativos16Bits ;Saltar si no es asi.
        and ah,ah                     ;Debe mostrarse el digito siempre?
        jnz PonerElDigitoEnBuffer16Bits ;Saltar si es asi.
        mov bh," "                    ;Mostrar espacio en vez de cero.
        jmp PonerElDigitoEnBuffer16Bits ;Poner el digito en buffer.
MostrarDigitoFormateado16:
        sub digitos16,1
        js short UltimaLineaFormateada16Bits
        mov di,PosicionDigitoEnBuffer
        mov cx,InformacionDeFormateado  ;CH = Numero de grupo (0-9).
                                        ;CL = Digito dentro del grupo (1-5).
        inc cx                          ;Indicar siguiente digito del grupo.
        cmp cl,6                        ;Fin del grupo?
        jnz short PonerDigitoFormateado16Bits  ;Saltar si no es asi.
        mov cl,1                        ;Indicar inicio del grupo.
        inc ch                          ;Indicar siguiente grupo.
        cmp ch,10                       ;Fin de la linea?
        jnz short PonerEspacio16Bits    ;Ir a mostrar espacio si no es asi.
        push ax         ;Preservar numero a convertir.
        push dx
        push bx         ;Preservar digito a poner en buffer.
        call MostrarBufferEnPantalla16Bits
        pop bx          ;BH = Digito en ASCII a poner en el buffer.
        mov di,dx       ;Borrar el buffer de digitos formateados.
        mov cx,75
        mov al," "
        cld
        rep stosb
        std
        inc cx
        mov di,dx       ;Apuntar al primer byte del buffer.
        inc di          ;Apuntar al siguiente caracter.
        pop dx          ;Restaurar el numero a convertir.
        pop ax
PonerEspacio16Bits:
        inc di          ;Apuntar al siguiente caracter.
PonerDigitoFormateado16Bits:
        mov byte ptr [di],bh    ;Copiar el digito en buffer de digitos form.
        inc di                  ;Incrementar el puntero.
        mov PosicionDigitoEnBuffer,di
        mov InformacionDeFormateado,cx
        inc DigitosMostrados    ;Indicar que se mostro un digito mas.
        ret                     ;Fin de la subrutina.
UltimaLineaFormateada16Bits:
        call MostrarBufferEnPantalla16Bits ;Mostrar la ultima linea en
                                           ;pantalla.
        mov ax,4c00h            ;Fin del programa.
        int 21h
;-------------------------------------------------------------------------
; Subrutina MostrarBufferEnPantalla16Bits:
MostrarBufferEnPantalla16Bits:
        mov di,offset BufferNumeroFormateado[66]
        or Indicadores,MostrarCantDigitos
        mov dx,DigitosMostrados
        mov al,0
        call Convertir6DigitosEnDI
        and Indicadores,not MostrarCantDigitos
        mov bx,handle
        mov dx,offset BufferNumeroFormateado
        mov cx,79
        mov ah,40h
        int 21h
        ret
;
;--------------------------------------------------------------------------
; Subrutina switch: Procesa los switches de linea de comandos inicializando
; los flags correspondientes.
;
; Entrada: BX = Puntero a la linea de comandos.
;
switch: cmp byte ptr [bx],"/"            ;Se encontro el caracter de switch?
        jz short SeEncontroSwitch        ;Saltar si es asi.
        cmp byte ptr [bx],ASCII_CR       ;Fin de la linea de comandos?
        jz short FinSwitch               ;No hay switches si es asi.
        inc bx                           ;Apuntar al sig. caracter de la linea.
        cmp byte ptr [bx-1]," "          ;Habia un espacio o caracter tab?
        jbe switch                       ;Saltearlo si es asi.
        lea dx,[bx-1]                    ;DX = Puntero al nombre del archivo.
buscar_fin_de_nombre_de_archivo:
        inc bx                           ;Apuntar al siguiente caracter.
        cmp byte ptr [bx-1]," "          ;Fin del nombre de archivo?
        ja buscar_fin_de_nombre_de_archivo ;Seguir buscando si no es asi.
        mov [bx-1],bh                    ;Convertir el nombre a ASCIIZ.
        mov ah,3Ch                       ;Intentar crear el archivo.
        xor cx,cx
        int 21h
        jc short JmpToError              ;Saltar si no se pudo.
        mov handle,ax                    ;Almacenar el handle.
FinSwitch:
        ret                              ;Fin de la subrutina.
JmpToError:
        jmp error                        ;Ir a la rutina de error.
SeEncontroSwitch:
        mov ax,[bx+1]                    ;Obtener el tipo de switch.
        and ax,0DFDFh                    ;Convertir a mayusculas.
        cmp ax,"NL"                      ;Es la primera parte de LN 2?
        jnz short switch2                ;Saltar si no es asi.
        cmp byte ptr [bx+3],"2"          ;Es la segunda parte de LN 2?
        jnz short switch2                ;Saltar si no es asi.
        or Indicadores,LN2               ;Indicar que se desea calcular ln 2.
        add bx,4                         ;Saltear el switch.
        jmp switch                       ;Ir a verificar si hay mas switches.
switch2:cmp al,"F"                       ;Es el switch de texto formateado?
        jnz JmpToError                   ;Ir a error si el switch es invalido.
        or Indicadores,Formato           ;Indicar que el texto debe formatearse
        add bx,2                         ;Saltear el switch.
        jmp switch                       ;Ir a verificar si hay mas switches.

;=============================================================================
; Calculo de PI y LN 2 con procesadores de 32 bits.
;=============================================================================

Inicio32:
        mov ax,4300h            ;Ver si hay un driver de XMS instalado en
        int 2Fh                 ;memoria.
        cmp al,80h
        jnz short HallarCantDecimales ;Saltar si no hay driver de XMS.
        mov ax,4310h            ;Obtener la direccion del manejador de XMS.
        int 2Fh
        mov word ptr [XMS],bx   ;Guardar esta direccion.
        mov word ptr [XMS+2],es
        push ds
        pop es                  ;Volver a poner ES = DS.
HallarCantDecimales:
        mov bx,80h              ;Apuntar al byte anterior del principio de
                                ;la linea de comandos.
        mov esi,10              ;ESI = Multiplicador.
        xor eax,eax             ;Inicializar numero a formar a cero.
        xor ecx,ecx             ;Inicializar digito a cero.
BusquedaDeDigito32:
        inc bx                  ;Apuntar al siguiente byte.
        mov al,[bx]             ;Obtener el caracter de la linea de comandos.
        cmp al,ASCII_CR         ;Se acabo la linea?
        je short error32        ;Saltar a error si es asi.
        cmp al," "              ;Es un espacio o caracter no imprimible?
        jbe BusquedaDeDigito32  ;Saltear dicho caracter.
        sub al,"0"              ;Ver si el caracter esta entre '0' y '9'.
        cmp al,9
        jbe short NumeroLineaDeComando32 ;Saltar si esta dentro del rango.
error32:                        ;Caso contrario ir a mostrar el error ya
        jmp error               ;que la linea debe comenzar con un numero.
NuevoDigito32:
        mul esi                 ;Multiplicar numero a formar por diez.
        jc short NoHaySuficienteMemoria ;Saltar si el resultado es muy grande.
        add eax,ecx             ;Sumar el digito hallado en la linea de comando
        jc short NoHaySuficienteMemoria ;Saltar si el resultado es muy grande.
NumeroLineaDeComando32:
        inc bx                  ;Apuntar al siguiente byte de la linea.
        mov cl,[bx]             ;Obtener el caracter.
        sub cl,"0"              ;Ver si esta entre '0' y '9'.
        cmp cl,9
        jbe NuevoDigito32       ;Sumar digito y repetir si esta en el rango.
        mov digitos32,eax       ;Almacenar cantidad de digitos en variable.
        cmp eax,39400           ;Limite para no usar memoria extendida.
        jbe short XMS_Opcional  ;Saltar si no es necesario el uso de XMS.
        cmp XMS_Instalado,0     ;Si el manejador de XMS no esta instalado,
        jz error32              ;es una condicion de error.
        jmp short HallarCantDWords ;Ir a calcular la cantidad de DWords.
XMS_Opcional:
        mov XMS_Instalado,0     ;Forzar indicacion de XMS no instalado.
                                ;De esta manera no se usara memoria extendida,
                                ;acelerando el programa para cant.dec.< 39400.
HallarCantDWords:
        push bx                 ;Preservar puntero a la linea de comando.
        mov esi,445861642       ;2^32 / log   (2^32)
        mul esi                 ;          10
        add edx,3               ;Reservar 3 DWords mas por el redondeo.
        mov CantDWords,edx      ;Almacenar la cantidad de DWords necesarias.
        cmp XMS_Instalado,0     ;El manejador de XMS esta instalado?
        jz short InicializarPrimerTermino  ;Saltar si no es asi.
        add edx,16384/4 - 1     ;Convertir la cantidad de DWords a cantidad
        shr edx,14 - 2          ;de bloques de 16KB.
        mov LongitudBufferx16KB,dx ;Preservar la cantidad de bloques de 16KB.
        shl edx,14 - 10         ;Convertir a cantidad de KB.
        lea edx,[edx * 2 + edx] ;Reservar espacio para 3 buffers.
        mov ah,9                ;Reservar esta cantidad de KB en
        call XMS                ;la memoria extendida.
        and ax,ax               ;Hubo errores?
        jnz short LosDigitosEntran ;Saltar si no hubo errores.
NoHaySuficienteMemoria:
        mov dx,offset MemoriaLlena ;Indicar la condicion de error.
        mov cx,LongitudMemoriaLlena
        mov ah,40h
        int 21h
        mov ax,4C02h            ;Saltar si no es asi.
        int 21h
LosDigitosEntran:
        mov XMS_Handle,dx       ;Preservar el handle de memoria extendida.
InicializarPrimerTermino:
        pop bx                  ;Restaurar puntero a la linea de comando.
;------------------------------------------------------------------------------
;
; Ver si en la linea de comandos aparecen /F o /LN2. En el caso que existan,
; se ponen a uno ciertos bits de la variable Indicadores. Para /F se pone a
; 1 el bit Formato, mientras que para /LN2 se pone a 1 el bit LN2.
;
        call switch
;
;------------------------------------------------------------------------------
;BORRAR CUANDO ESTE LISTO LN 2 PARA 32 BITS
        test Indicadores,LN2    ;Calcular logaritmo natural de 2?
        jz short Pi32Bits       ;Saltar si no es asi.
        cmp XMS_Instalado,0     ;Se uso XMS?
        jz Inicio16             ;Saltar si no es asi.
        mov dx,XMS_Handle
        mov ah,0Ah              ;Liberar el handle de XMS.
        call XMS
        jmp Inicio16            ;Calcular LN 2 con codigo de 16 bits.
Pi32Bits:
;------------------------------------------------------------------------------
        call HabilitarInterrupcion
        mov edx,13753*4         ;Primer numerador.
        mov ecx,CantDWords      ;Obtener cantidad de DWords del numero.
        cmp XMS_Instalado,0     ;Se utilizara memoria extendida?
        jz CalculoPrimerTermRAMConv ;Saltar si no es asi.
        xor ebx,ebx             ;Apuntar al principio del numero ubicado en
                                ;memoria extendida.
        mov ebp,ecx             ;Convertir cantidad de DWords a cantidad
        shr ebp,14 - 2          ;de bloques de 16 KB.
        and cx,0FFFh            ;CX = Cantidad de DWords en el ultimo bloque.
        push cx                 ;Preservar este valor.
        jz short CicloXMSCalcPrimerTermino ;Saltar si no hay DWords
                                ;en el ultimo bloque.
        and bp,bp
        jz short CalcularUltimoBloquePrimerTermino
CicloXMSCalcPrimerTermino:
        mov cx,1000h            ;Indicar cantidad de DWords por bloque.
        push bp                 ;Preservar la cantidad de bloques que faltan.
        call SubrCicloXMSCalcPrimerTermino ;Calcular en este bloque el
                                ;primer termino.
        add ebx,4000h           ;Apuntar al siguiente bloque.
        pop bp                  ;Restaurar la cantidad de bloques que faltan.
TerminarCicloXMSCalcPrimerTermino:
        dec bp                  ;Decrementar la cantidad de bloques que faltan.
        jnz CicloXMSCalcPrimerTermino ;Repetir si faltan calcular bloques.
CalcularUltimoBloquePrimerTermino:
        pop cx                  ;Restaurar la cantidad de DWords en el ultimo
                                ;bloque.
        jcxz CicloPrincipalXMS32Bits
        call SubrCicloXMSCalcPrimerTermino ;Calcular en este bloque el
                                ;primer termino.
        push ax                 ;Preparar para siguiente instruccion.
CicloPrincipalXMS32BitsPop:
        pop ax                  ;Descartar lo que entro en la pila.
CicloPrincipalXMS32Bits:
        call NuevoTerminoXMS32  ;Calcular el siguiente termino.
        pushf                   ;Preservar el indicador de arrastre (CF).
CicloRestaDeBloquesXMS:
        jcxz ObtenerProxBloqueXMSEnResta
        popf                    ;Restaurar el indicador de arrastre (CF).
CicloRestaDelTerminoXMS:
        mov eax,[di]            ;Obtener DWord de la suma.
        sbb eax,[di+16384]      ;Restar DWord del termino incluyendo el
                                ;prestamo del DWord anterior.
        stosd                   ;Guardar resta actualizando el puntero.
        loop CicloRestaDelTerminoXMS ;Seguir restando hasta terminar el bloque.
        pushf                   ;Preservar el indicador de arrastre (CF).
ObtenerProxBloqueXMSEnResta:
        movzx eax,LongitudBufferx16KB ;EAX = Longitud del numero en bloques
                                ;de 16 KB.
        shl eax,14 + 1          ;EAX = Dos veces la longitud del numero.
        sub ebx,eax             ;Apuntar al numero suma.
        and edx,edx             ;Se termino el numero (ultimo bloque)?
        jz short SeguirRestandoXMS ;Saltar si es asi.
        call ObtenerSiguienteBloqueXMS  ;Obtener el siguiente bloque.
        jmp CicloRestaDeBloquesXMS  ;Seguir realizando la resta.
SeguirRestandoXMS:
        sub di,offset Buffer
        jnc short FinRestaEnMedioDeBloque
        popf
FinBloqueEnResta:
        pushf
        mov OffsetFuenteRAMConv,offset Buffer
        call SubrEnviarDatosHaciaXMS
        popf
        jnc short CalcularNuevoTerminoXMS
        pushf
        mov CantidadBytesAMover,16384
        sub ebx,16384
        mov OffsetDestinoRAMConv,offset Buffer
        call SubrEnviarDatosDesdeXMS
        popf
        jnc short CalcularNuevoTerminoXMS
        mov di,16380
CicloSeguirRestandoXMS:
        sub di,4                ;Apuntar al siguiente DWord mas significativo.
        jc short FinBloqueEnResta
        sub dword ptr Buffer[di+4],1 ;Decrementar el DWord.
        jc short CicloSeguirRestandoXMS ;No continuar el ciclo si los DWords
                                ;mas significativos no se deben modificar.
        and di,di
        jz FinBloqueEnResta
        pushf
FinRestaEnMedioDeBloque:
        popf
        jc short CicloSeguirRestandoXMS
        mov OffsetFuenteRAMConv,offset Buffer
        call SubrEnviarDatosHaciaXMS
CalcularNuevoTerminoXMS:
        call NuevoTerminoXMS32
        pushf
CicloSumaDeBloquesXMS:
        jcxz ObtenerProxBloqueXMSEnSuma
        popf                    ;Restaurar el indicador de arrastre (CF).
CicloSumaDelTerminoXMS:
        mov eax,[di]            ;Obtener DWord de la suma.
        adc eax,[di+16384]      ;Sumar DWord del termino incluyendo el
                                ;acarreo del DWord anterior.
        stosd                   ;Guardar suma actualizando el puntero.
        loop CicloSumaDelTerminoXMS ;Seguir sumando hasta terminar el bloque.
        pushf                   ;Preservar el indicador de arrastre (CF).
ObtenerProxBloqueXMSEnSuma:
        movzx eax,LongitudBufferx16KB ;EAX = Longitud del numero en bloques
                                ;de 16 KB.
        shl eax,14 + 1          ;EAX = Dos veces la longitud del numero.
        sub ebx,eax             ;Apuntar al numero suma.
        and edx,edx             ;Se termino el numero (ultimo bloque)?
        jz short SeguirSumandoXMS ;Saltar si es asi.
        call ObtenerSiguienteBloqueXMS  ;Obtener el siguiente bloque.
        jmp CicloSumaDeBloquesXMS  ;Seguir realizando la suma.
SeguirSumandoXMS:
        sub di,offset Buffer
        jnc short FinSumaEnMedioDeBloque
        popf
FinBloqueEnSuma:
        pushf
        mov OffsetFuenteRAMConv,offset Buffer
        call SubrEnviarDatosHaciaXMS
        popf
        jnc CicloPrincipalXMS32Bits
        pushf
        mov CantidadBytesAMover,16384
        sub ebx,16384
        mov OffsetDestinoRAMConv,offset Buffer
        call SubrEnviarDatosDesdeXMS
        popf
        jnc CicloPrincipalXMS32Bits
        mov di,16380
CicloSeguirSumandoXMS:
        sub di,4                ;Apuntar al siguiente DWord mas significativo.
        jc short FinBloqueEnSuma
        add dword ptr Buffer[di+4],1 ;Decrementar el DWord.
        jc short CicloSeguirSumandoXMS ;No continuar el ciclo si los DWords
                                ;mas significativos no se deben modificar.
        and di,di
        jz FinBloqueEnSuma
        pushf
FinSumaEnMedioDeBloque:
        popf
        jc short CicloSeguirSumandoXMS
        mov OffsetFuenteRAMConv,offset Buffer
        call SubrEnviarDatosHaciaXMS
        jmp CicloPrincipalXMS32Bits
CalculoPrimerTermRAMConv:
        mov ebx,63664921        ;Primer denominador.
        mov di,offset buffer
CicloCalcPrimerTermino:
        xor eax,eax
        div ebx
        mov [di+16384],eax
        stosd
        loop CicloCalcPrimerTermino
        mov byte ptr buffer[3],256*4/32+256*4/128
        mov byte ptr buffer[16384+3],256*4/32+256*4/128
CicloPrincipal32:
        call NuevoTermino32Bits
CicloRestaDelTermino32:
        mov eax,[di]
        sbb eax,[si]
        stosd
        lea si,[si-4]
        loop CicloRestaDelTermino32
CicloSeguirRestando32:
        sbb dword ptr [di],0
        lea di,[di-4]
        jc CicloSeguirRestando32
        call NuevoTermino32Bits
CicloSumaDelTermino32:
        lodsd
        adc eax,[di]
        stosd
        loop CicloSumaDelTermino32
CicloSeguirSumando32:
        adc dword ptr [di],0
        lea di,[di-4]
        jc CicloSeguirSumando32
        jmp CicloPrincipal32
;--------------------------------------------------------------------------
PonerPotenciasBinarias32Bits:
SacarPotenciasBinarias32Bits:
        mov edi,divisor32
        lea edi,[edi*2+edi]
        sub edi,5
        jb short OtrasPotencias32Bits
        mov cx,di
        mov eax,0C0000000h
        ror eax,cl
        shr edi,3
        and di,0FFFCh
        add di,offset buffer[16384]
        xor [di],eax
OtrasPotencias32bits:
        mov eax,5
        call Potencia32Bits
        mov eax,7
Potencia32Bits:
        mul divisor32
        sub eax,2
        mov cl,al
        inc edx
        ror edx,cl
        shr eax,5
        cmp ax,CantWords
        jae short FinSubrutina32
        shl ax,2
        mov si,ax
        xor dword ptr buffer[si+16384],edx
FinSubrutina32:
        ret
;------------------------------------------------------------------------------
; Subrutinas PonerPotenciasBinarias32BitsXMS y SacarPotenciasBinarias32BitsXMS:
; Mediante un "xor", pone o saca las potencias de dos que figuran en la
; formula que figura al principio del programa.
;
PonerPotenciasBinarias32BitsXMS:
SacarPotenciasBinarias32BitsXMS:
;
; Hay que calcular 24 / 8 ^ (div), donde "div" es el divisor (3, 5, 7,..)
; del termino. Para ello hay que obtener la posicion de los bits correspon-
; dientes a "xorear" de la siguiente manera:
;
;          24 / 8 ^ (div) = 3 / 8 ^ (div - 1)
;                         = 3 / 2 ^ (3 * div - 3)
;                         = 2 ^ -(3 * div - 4) + 2 ^ -(3 * div - 3)
;
; Por lo tanto los bits a "xorear" son los correspondientes a las potencias
; negativas de dos: 3 * div - 4 y 3 * div - 3.
;
; Ahora bien, como solo trabajamos con la parte decimal de los numeros, los
; bits se numeran como bit 0 -> 2^(-1), bit 1 -> 2^(-2), etc., las posiciones
; como numero de bit seran 3 * div - 5 y 3 * div - 4.
;
        mov edi,divisor32           ;EDI <- div
        lea edi,[edi*2+edi]         ;EDI <- 3 * div
        sub edi,5                   ;EDI <- 3 * div - 5
        jb short SaltearPotenciasUnOctavo ;Si hay carry es porque el divisor
                                    ;es 1. En este caso 24 / 8^div es un
                                    ;numero entero, por lo que no hay que
                                    ;hacer nada.
        push edi                    ;Preservar posicion de bit.
        call CambiarBitEDIEnXMS     ;Cambiar el estado del bit.
        pop edi                     ;Restaurar posicion de bit.
        inc edi                     ;Procesar siguiente bit menos significativo
        call CambiarBitEDIEnXMS     ;Cambiar el estado del bit.
SaltearPotenciasUnOctavo:
        mov eax,5                   ;Procesar potencias de 32 (2 ^ 5).
        call Potencia32BitsXMS
        mov eax,7                   ;Procesar potencias de 128 (2 ^ 7).
Potencia32BitsXMS:
;
; Debemos hallar ahora la posicion de bit correspondiente a (ver formula):
;
;       4 / (2 ^ exp) ^ div = 4 / 2 ^ (exp * div)
;                           = 1 / 2 ^ (exp * div - 2)
;
; Esto significa que hay que cambiar el bit numero exp * div - 3.
;
        mul divisor32               ;EAX = exp * div
        lea edi,[eax - 3]           ;EDI = exp * div - 3
CambiarBitEDIEnXMS:
        mov eax,edi                 ;EAX = Posicion de bit.
        shr edi,5                   ;EDI = Convertir numero de bit a DWord
                                    ;      (1 DWord = 32 bits).
        cmp edi,CantDWords          ;Numero de bit fuera de rango?
        jnc short FinPonerBitEDIEnXMS ;No cambiar el bit si es asi.
        push ax                     ;Preservar posicion de bit dentro del
                                    ;DWord (en los 5 bits menos signif. de AX).
        mov CantidadBytesAMover,4   ;Indicar que se va a traer una DWord desde
                                    ;memoria extendida.
        mov ax,XMS_Handle
        mov HandleXMSFuente,ax      ;La fuente es la memoria extendida.
        movzx eax,LongitudBufferx16KB ;EAX = Longitud del numero en bloques
                                    ;de 16 KB.
        shl eax,14                  ;EAX = Longitud del numero en bytes.
        lea eax,[EDI * 4 + EAX]     ;EAX = Posicion del byte a traer dentro
                                    ;del segundo numero (el termino).
        mov OffsetXMSFuente,eax     ;Llenar la estructura de copia de XMS.
        mov HandleXMSDestino,0
        mov OffsetDestinoRAMConv,offset Buffer ;Poner los cuatro bytes a
        mov SegmentoDestinoRAMConv,ds          ;partir de "Buffer".
        call CopiarBytesXMS
        pop cx                      ;Obtener posicion de bit dentro del
                                    ;DWord (en los 5 bits menos signif. de CL).
        mov eax,80000000h           ;Encender el bit 0 (contando desde MSB).
        mov HandleXMSFuente,ax      ;La fuente sera memoria RAM convencional.
        shr eax,cl                  ;Encender el bit que corresponde dentro
                                    ;del DWord (mascara).
        xor dword ptr Buffer,eax    ;Cambiar el bit del DWord.
        mov ax,XMS_Handle
        mov HandleXMSDestino,ax     ;El destino es la memoria extendida.
        mov eax,OffsetXMSFuente     ;Intercambiar offset de fuente y destino
        xchg eax,OffsetXMSDestino   ;(porque la copia es en sentido inverso).
        mov OffsetXMSFuente,eax
CopiarBytesXMS:
        mov si,offset EstructuraCopiarBytesXMS
        mov ah,0Bh                  ;Funcion para copiar memoria extendida.
        call XMS                    ;Realizar la copia.
FinPonerBitEDIEnXMS:
        ret                         ;Fin de la subrutina.
;------------------------------------------------------------------------------
; Subrutina ObtenerSiguienteBloqueXMS:
;
; Entrada: EBX = Puntero al bloque de memoria extendida a obtener.
;          EDX = Cantidad de DWords que faltan procesar.
;
; Salida: EDX = Cantidad de DWords que faltan procesar.
;         ECX = Cantidad de DWords a procesar en el bloque de XMS.
;         DI = Posicion de la DWord a sumar o restar.
;
ObtenerSiguienteBloqueXMS:
        mov OffsetFuenteRAMConv,offset Buffer  ;Salvar el buffer de resultados
        call SubrEnviarDatosHaciaXMS           ;en memoria extendida.
        mov ecx,1000h                   ;Cantidad de DWords por bloque.
        sub edx,ecx                     ;Restar la cantidad de DWords.
        jnc short BytesEnElBloqueOK     ;Saltar si no es menor que cero.
        add ecx,edx                     ;Obtener la cantidad real de DWords
                                        ;de este bloque.
        xor edx,edx                     ;Indicar ultimo bloque.
BytesEnElBloqueOK:
        shl cx,2                        ;Convertir a cantidad de bytes.
        push cx
        mov CantidadBytesAMover,cx      ;Almacenarlo en estructura de copia.
        mov OffsetDestinoRAMConv,offset Buffer ;Cargar el siguiente bloque
        sub ebx,ecx                            ;mas significativo del resultado
        call SubrEnviarDatosDesdeXMS           ;en el area de trabajo.
        movzx eax,LongitudBufferx16KB   ;EAX = Cantidad de bloques del numero.
        shl eax,14 + 1                  ;EAX = Cantidad de bytes de dos nros.
        add ebx,eax                     ;Apuntar al bloque del termino en XMS.
        mov OffsetDestinoRAMConv,offset Buffer[16384] ;Cargar el siguiente
                                        ;bloque XMS del termino.
        call SubrEnviarDatosDesdeXMS    ;Obtener el bloque del termino.
        pop di
        mov cx,di
        shr cx,2
        add di,offset buffer[-4]        ;Ultima DWord del bloque del numero.
        ret                             ;Fin de la subrutina.
;------------------------------------------------------------------------------
; Subrutina SubrCicloXMSCalcPrimerTermino: Como el numero almacenado en memoria
; extendida es muy extenso, el calculo debe realizarse en bloques de 16 KB.
; El proposito de esta subrutina es calcular un bloque del primer termino.
;
; Entrada: EBX = Puntero al bloque en memoria extendida. La rutina llamadora
;                no debe alterar EBX entre sucesivas llamadas a esta subrutina.
;          EDX = Resto de la ultima division. La rutina llamadora no debe
;                alterar EDX entre sucesivas llamadas a esta subrutina.
;
SubrCicloXMSCalcPrimerTermino:
        push ebx                ;Preservar el puntero a memoria extendida.
        mov OffsetXMSDestino,ebx ;Inicializar puntero a memoria extendida en
                                ;estructura para copiar memoria.
        mov ebx,63664921        ;Primer denominador.
        mov di,offset buffer    ;Puntero al area de trabajo en RAM convencional
        mov CantidadBytesAMover,cx ;Rellenar estructura para realizar la copia.
        shl CantidadBytesAMover,2  ;Convertir cantidad de DWords a bytes.
CicloXMSCalcPrimerTerminoEnBuffer:
        xor eax,eax             ;La parte decimal del denominador es cero
                                ;en el primer termino.
        div ebx                 ;Generar DWord del cociente.
        stosd                   ;Almacenarlo en el area de trabajo.
        loop CicloXMSCalcPrimerTerminoEnBuffer ;Cerrar el ciclo.
        pop ebx                 ;Restaurar el puntero a memoria extendida.
        and ebx,ebx             ;Apunta al principio?
        jnz short CopiarBufferEnXMS ;Saltar si no es asi.
        mov byte ptr buffer[3],256*4/32+256*4/128 ;Agregar potencias de 2.
CopiarBufferEnXMS:
        mov OffsetFuenteRAMConv,offset buffer;Rellenar estructura para realizar
                                             ;la copia.
        call SubrEnviarDatosHaciaXMS ;Copiar area de trabajo en mem. extendida.
        push edx                ;Preservar resto de la ultima division.
        push ebx                ;Preservar puntero a memoria extendida.
        movzx eax,LongitudBufferx16KB
        shl eax,14              ;Convertir bloques de 16 KB a bytes.
        add OffsetXMSDestino,eax;Indicar donde se debe copiar el area de
                                ;trabajo en la estructura de control de copia.
        call CopiarBytesXMS     ;Realizar la copia hacia memoria extendida.
        pop ebx                 ;Restaurar puntero a memoria extendida.
        pop edx                 ;Restaurar resto de la ultima division.
        ret                     ;Final de la subrutina.
;------------------------------------------------------------------------------
; Subrutina SubrEnviarDatosDesdeXMS: Sirve para copiar un bloque de memoria
; ubicado en memoria extendida a un area de trabajo ubicado en memoria
; convencional (debajo de 1MB).
;
; Entrada: EBX = Posicion en memoria extendida del bloque a copiar.
; OffsetDestinoRAMConv = Posicion en el segmento de datos donde se ubica el
;                        area de trabajo.
; La subrutina preserva EBX y EDX.
SubrEnviarDatosDesdeXMS:
        push edx                ;Preservar EDX.
        push ebx                ;Preservar el puntero a memoria extendida.
        mov ax,XMS_Handle       ;Obtener el handle de memoria extendida.
        mov HandleXMSFuente,ax  ;Indicar que se va a copiar desde el area de
                                ;memoria extendida pedida por este programa.
        mov OffsetXMSFuente,ebx ;Indicar desde donde se va a copiar.
        mov HandleXMSDestino,0  ;Indicar que el destino esta por debajo de 1MB.
        mov SegmentoDestinoRAMConv,ds ;Indicar el segmento destino.
        call CopiarBytesXMS     ;Realizar la copia desde memoria extendida.
        pop ebx                 ;Restaurar el puntero a memoria extendida.
        pop edx                 ;Restaurar EDX.
        ret                     ;Fin de la subrutina.
;------------------------------------------------------------------------------
; Subrutina SubrEnviarDatosHaciaXMS: Sirve para copiar un area de trabajo
; ubicado en memoria convencional (debajo de 1MB) a una zona de memoria
; extendida.
;
; Entrada: EBX = Posicion en memoria extendida del bloque donde se copiara
;                el area de trabajo.
; OffsetFuenteRAMConv = Posicion en el segmento de datos donde se ubica el
;                       area de trabajo.
; La subrutina preserva EBX y EDX.
SubrEnviarDatosHaciaXMS:
        push edx                ;Preservar EDX.
        push ebx                ;Preservar el puntero a memoria extendida.
        mov ax,XMS_Handle       ;Obtener el handle de memoria extendida.
        mov HandleXMSDestino,ax ;Indicar que se va a copiar hacia el area de
                                ;memoria extendida pedida por este programa.
        mov OffsetXMSDestino,ebx;Indicar hacia donde se va a copiar.
        mov HandleXMSFuente,0   ;Indicar que la fuente esta por debajo de 1MB.
        mov SegmentoFuenteRAMConv,ds ;Indicar el segmento fuente.
        call CopiarBytesXMS     ;Realizar la copia hacia memoria extendida.
        pop ebx                 ;Restaurar el puntero a memoria extendida.
        pop edx                 ;Restaurar EDX.
        ret                     ;Fin de la subrutina.
;------------------------------------------------------------------------------
NuevoTerminoXMS32:
        call SacarPotenciasBinarias32BitsXMS ;Retirar potencias de 2 del
                                       ;termino anterior.
        mov ebx,PrimDWord              ;EBX = Palabra mas significativa
                                       ;distinta de ceros.
        mov eax,CantDWords             ;EAX = Cantidad de palabras del numero.
        sub eax,ebx                    ;EAX = Cantidad de palabras a procesar.
        jbe short CantidadPalabXMSDelTerminoOk32   ;Saltar si es <= 0.
        mov cx,ax                      ;CX = Cantidad de DWords en el
        and cx,0FFFh                   ;     ultimo bloque de 16 KB.
        shl ebx,2                      ;EBX = Puntero al primer byte.
        sub ebx,4000h                  ;Ajustar el offset (la subrutina
                                       ;arranca sumando este valor).
        movzx edx,LongitudBufferx16KB  ;EDX = Longitud del numero en bloques
                                       ;de 16 KB (para apuntar al 2do numero).
        shl edx,14                     ;EDX = Longitud del numero en bytes.
        add ebx,edx                    ;EBX = Apunta ahora al segundo numero.
        mov OffsetXMSDestino,ebx
        xor edx,edx                    ;Inicializar resto de la 1ra division.
        xor ebx,ebx                    ;Inicializar resto de la 2do division.
        mov ebp,eax                    ;EBP = Cantidad de DWords a procesar.
        shr ebp,14 - 2                 ;BP = Cantidad de bloques completos.
        push bp                        ;Preservar la cantidad de bloques.
        jz short DividirUltimoBloqueXMS ;Saltar si no hay un bloque completo.
        push cx                        ;Preservar cantidad de DWords en el
                                       ;ultimo bloque.
CicloDividirBloqueXMS:
        mov cx,1000h                   ;Cantidad de DWords en el bloque.
        push bp                        ;Preservar contador de bloques de XMS.
        call SubrDividirBloqueXMS      ;Realizar la division.
        pop bp                         ;Restaurar contador de bloques de XMS.
        dec bp                         ;Decrementar contador de bloques.
        jnz CicloDividirBloqueXMS      ;Cerrar el ciclo.
        pop cx                         ;Restaurar cantidad de DWords en el
                                       ;ultimo bloque.
DividirUltimoBloqueXMS:
        jcxz UltimaDivisionHecha
        push cx                        ;Preservar cantidad de DWords en el
                                       ;ultimo bloque.
        call SubrDividirBloqueXMS      ;Realizar la division en ultimo bloque.
        pop cx                         ;Restaurar cantidad de DWords en el
                                       ;ultimo bloque.
UltimaDivisionHecha:
        mov ebx,13753*13753/2          ;Inicializar arrastre de DWord menos
                                       ;significativo.
        mov esi,13753*13753            ;Numerador.
        inc bp
        jcxz PrimeraMultHecha
        call SubrMultiplicarBloqueXMS  ;Realizar la multiplicacion en el
                                       ;ultimo bloque.
PrimeraMultHecha:
        pop bp                         ;Restaurar la cantidad de bloques
                                       ;completos.
        and bp,bp                      ;Hay algun bloque completo?
        jz short FinMultiplicacionXMS  ;Saltar si no es asi.
CicloMultiplicarBloqueXMS:
        mov cx,1000h                   ;Cantidad de DWords en el bloque.
        push bp                        ;Preservar contador de bloques de XMS.
        mov esi,13753*13753            ;Numerador.
        call SubrMultiplicarBloqueXMS  ;Realizar la multiplicacion.
        pop bp                         ;Restaurar contador de bloques de XMS.
        dec bp                         ;Decrementar contador de bloques.
        jnz CicloMultiplicarBloqueXMS  ;Cerrar el ciclo.
FinMultiplicacionXMS:
        cmp dword ptr buffer,0         ;Se hizo cero el DWord mas significat?
        jnz short CantidadPalabXMSDelTerminoOk32 ;Saltar si no es asi.
        inc PrimDWord                  ;Incrementar indice de DWord mas
                                       ;significativo.
CantidadPalabXMSDelTerminoOk32:
        add divisor32,2                ;Siguiente divisor.
        call PonerPotenciasBinarias32BitsXMS ;Poner las potencias binarias.
        mov ebx,divisor32              ;EBX <- div
        lea ebx,[ebx*2+ebx-4]          ;EBX <- 3 * div - 4
        shr ebx,5                      ;EBX = Numero de DWord.
        mov ecx,CantDWords             ;ECX = Cantidad de Dwords en el numero.
        sub ecx,ebx                    ;ECX = Cantidad de Dwords a procesar.
        jbe BinarioADecimal32          ;Si no hay DWords para dividir, se acabo
                                       ;el calculo. Hay que pasar de binario a
                                       ;decimal.
        mov ebp,ecx                    ;EBP = Cantidad de DWords.
        shl ebx,2                      ;EBX = Puntero al primer byte.
        and cx,0FFFh                   ;CX = Cantidad de DWords en el
                                       ;     ultimo bloque de 16 KB.
        movzx eax,LongitudBufferx16KB  ;EAX = Longitud del numero en bloques
                                       ;de 16 KB.
        shl eax,14                     ;EAX = Longitud del numero en bytes.
        lea ebx,[ebx+eax-4000h]        ;Ajustar el offset (la subrutina
                                       ;arranca sumando este valor).
        mov OffsetXMSDestino,ebx
        xor edx,edx                    ;Inicializar resto de la division.
        shr ebp,14 - 2                 ;BP = Cantidad de bloques completos.
        push bp                        ;Preservar la cantidad de bloques.
        jz short DivisionVariableUltimoBloqueXMS  ;Saltar si no hay un bloque
                                                  ;completo.
        push cx                        ;Preservar cantidad de DWords en el
                                       ;ultimo bloque.
CicloDivisionVariableBloqueXMS:
        mov cx,1000h                   ;Cantidad de DWords en el bloque.
        push bp                        ;Preservar contador de bloques de XMS.
        call SubrDivisionVariableBloqueXMS ;Realizar la division.
        pop bp                         ;Restaurar contador de bloques de XMS.
        dec bp                         ;Decrementar contador de bloques.
        jnz CicloDivisionVariableBloqueXMS ;Cerrar el ciclo.
        pop cx                         ;Restaurar cantidad de DWords en el
                                       ;ultimo bloque.
DivisionVariableUltimoBloqueXMS:
        jcxz UltimaDivVariableHecha
        push cx                        ;Preservar cantidad de DWords en el
                                       ;ultimo bloque.
        call SubrDivisionVariableBloqueXMS ;Realizar la division en ultimo
                                           ;bloque.
        pop cx                         ;Restaurar cantidad de DWords en el
                                       ;ultimo bloque.
UltimaDivVariableHecha:
        pop bp                         ;Restaurar la cantidad de bloques
                                       ;completos.
        and cx,cx
        jnz short BloqueNoNulo
        dec bp
        mov cx,1000h
BloqueNoNulo:
        movzx edx,bp                   ;EDX = Cantidad de bloques completos.
        shl edx,14 - 2                 ;EDX = Cantidad de DWords.
        movzx ebx,LongitudBufferx16KB
        shl ebx,14
        add ebx,OffsetXMSDestino
        mov OffsetDestinoRAMConv,offset buffer[16384]
        shl cx,2
        push cx
        mov CantidadBytesAMover,cx
        call SubrEnviarDatosDesdeXMS   ;Enviar bloque al area de trabajo.
        push ebx                       ;Preservar el puntero a XMS.
        movzx eax,LongitudBufferx16KB  ;EAX = Cantidad de bloques del numero.
        shl eax,14 + 1                 ;EAX = Cantidad de bytes del numero.
        sub ebx,eax                    ;Apuntar al bloque del termino en XMS.
        mov OffsetDestinoRAMConv,offset buffer
        call SubrEnviarDatosDesdeXMS   ;Obtener el bloque del termino.
        pop ebx                        ;Restaurar el puntero.
        pop cx                         ;Restaurar cantidad de DWords en bloque.
        mov di,offset buffer[-4]       ;Apuntar al area de trabajo.
        add di,cx                      ;Apuntar al DWord menos significativos.
        shr cx,2                       ;Obtener nuevamente cantidad de DWords.
        std
        ret
;----------------------------------------------------------------------------
SubrDividirBloqueXMS:
        push ebx                       ;Preservar resto de la segunda division.
        mov ebx,OffsetXMSDestino
        add ebx,4000h                  ;Apuntar al siguiente bloque de XMS.
        mov OffsetDestinoRAMConv,offset buffer
        push cx                        ;Preservar cantidad de DWords en bloque.
        shl cx,2
        mov CantidadBytesAMover,cx
        call SubrEnviarDatosDesdeXMS   ;Enviar bloque al area de trabajo.
        pop cx                         ;Restaurar cantidad de DWords en bloque.
        pop ebx                        ;Restaurar resto de la segunda division.
        mov di,offset buffer           ;Apuntar al area de trabajo.
        mov esi,63664921               ;Cargar divisor.
        cld                            ;Procesar bloque en forma ascendente.
CicloDivXMSPor63664921ALa2:
        mov eax,[di]
        div esi
        xchg ebx,edx
        div esi
        xchg ebx,edx
        stosd
        loop CicloDivXMSPor63664921ALa2
        push ebx
        mov ebx,OffsetXMSFuente
        mov OffsetFuenteRAMConv,offset buffer
        call SubrEnviarDatosHaciaXMS
        pop ebx
        ret
;------------------------------------------------------------------------------
; Subrutina SubrDivisionVariableBloqueXMS: Divide un bloque de memoria XMS por
; la variable DWord divisor32.
; Entrada: EDX = Resto.
;          CX  = Cantidad de DWords en el bloque.
; Salida:  EDX = Resto.
SubrDivisionVariableBloqueXMS:
        mov ebx,OffsetXMSDestino
        add ebx,4000h
        mov OffsetDestinoRAMConv,offset buffer
        push cx                        ;Preservar cantidad de DWords en bloque.
        shl cx,2                       ;Convertir a cantidad de bytes.
        mov CantidadBytesAMover,cx
        call SubrEnviarDatosDesdeXMS   ;Enviar bloque al area de trabajo.
        pop cx                         ;Restaurar cantidad de DWords en bloque.
        mov di,offset buffer           ;Apuntar al area de trabajo.
        mov si,di
        mov ebx,divisor32              ;Cargar divisor.
        cld                            ;Procesar bloque en forma ascendente.
CicloDivXMSPorNroVariable:
        lodsd
        div ebx
        stosd
        loop CicloDivXMSPorNroVariable
        movzx ebx,LongitudBufferx16KB  ;EBX = Longitud del numero en bloques
                                       ;de 16 KB.
        shl ebx,14                     ;Convertir a bytes.
        add ebx,OffsetXMSFuente
        mov OffsetFuenteRAMConv,offset buffer
        call SubrEnviarDatosHaciaXMS
        movzx ebx,LongitudBufferx16KB  ;EBX = Longitud del numero en bloques
                                       ;de 16 KB.
        shl ebx,14                     ;Convertir a bytes.
        sub OffsetXMSDestino,ebx
        ret
;------------------------------------------------------------------------------
; Subrutina SubrMultiplicarBloqueXMS: Multiplica un bloque de XMS del
; numero por un DWord.
; Entrada: ESI = DWord a usar como factor.
;          EBX = Arrastre de la multiplicacion del bloque anterior.
;          CX = Cantidad de DWords en el bloque de XMS.
SubrMultiplicarBloqueXMS:
        push ebx
        push esi
        mov ebx,OffsetXMSDestino
        mov OffsetDestinoRAMConv,offset buffer
        shl cx,2
        push ecx                      ;Preservar cantidad de bytes en bloque.
        mov CantidadBytesAMover,cx
        call SubrEnviarDatosDesdeXMS
        pop ecx
        mov di,offset buffer[-4]
        add di,cx
        shr cx,2
        pop esi
        pop ebx
        std
CicloMultXMSPorESI:
        mov eax,[di]
        mul esi
        add eax,ebx
        adc edx,0
        mov ebx,edx
        stosd
        loop CicloMultXMSPorESI
        and bp,bp
        jz short FinMultiplicarBloqueXMS
        push ebx
        mov ebx,OffsetXMSFuente
        mov OffsetFuenteRAMConv,offset buffer
        call SubrEnviarDatosHaciaXMS
        pop ebx
        sub OffsetXMSDestino,4000h
FinMultiplicarBloqueXMS:
        ret
;------------------------------------------------------------------------------
NuevoTermino32Bits:
        call SacarPotenciasBinarias32Bits  ;Retirar potencias de 2 del
                                       ;termino anterior.
        mov di,offset buffer[16384]    ;DI = Puntero al termino.
        mov ax,PrimWord                ;AX = Numero de palabra mas
                                       ;significativa distinta de ceros.
        mov cx,CantWords               ;CX = Cantidad de palabras del numero.
        sub cx,ax                      ;CX = Cantidad de palabras a procesar.
        jbe short CantidadPalabrasDelTerminoOk32    ;Saltar si es <= 0.
        shl ax,2                       ;AX = Indice dentro del puntero.
        add di,ax                      ;DI = Puntero a la palabra mas
                                       ;significativa distinta de ceros.
        mov esi,63664921               ;ESI = Divisor.
        xor edx,edx                    ;EDX = Resto de la primera division
                                       ;por 63664921.
        xor ebx,ebx                    ;EBX = Resto de la segunda division
                                       ;por 63664921.
        cld                            ;Procesar en direcciones ascendentes
                                       ;(para realizar una division se deben
                                       ;tomar primero los Dwords mas signif.)
        push cx                        ;Preservar la cantidad de palabras a
                                       ;procesar.
CicloDiv32BitsPor63664921ALa2:
        mov eax,[di]                   ;Obtener el DWord del dividendo.
        div esi                        ;Realizar la primera division.
        xchg ebx,edx                   ;EDX = Resto de la segunda division
                                       ;anterior.
                                       ;EBX = Resto de la primera division
                                       ;actual.
        div esi                        ;Realizar la segunda division.
        xchg ebx,edx                   ;EDX = Resto de la primera division
                                       ;actual.
                                       ;EBX = Resto de la segunda division
                                       ;actual.
        stosd                          ;Almacenar el cociente.
        loop CicloDiv32BitsPor63664921ALa2  ;Cerrar el ciclo.
        pop cx                         ;Restaurar la cantidad de palabras a
                                       ;procesar.
        std                            ;Procesar en direcciones descendentes
                                       ;(para realizar una multiplicacion se
                                       ;deben tomar primero los Dwords menos
                                       ;significativos).
        sub di,4                       ;Apuntar al DWord menos significativo.
        mov ebx,13753*13753/2          ;EBX = Multplicador / 2 (para redondeo)
        mov esi,13753*13753            ;ESI = Multiplicador.
CicloMult32BitsPor13753ALa2:
        mov eax,[di]                   ;Obtener DWord del factor.
        mul esi                        ;Realizar la multiplicacion.
        add eax,ebx                    ;Sumar acarreo de multiplic. anterior.
        adc edx,0                      ;Ajustar el DWord mas significativo.
        mov ebx,edx                    ;Guardar acarreo de la multiplicacion.
        stosd                          ;Almacenar DWord del producto.
        loop CicloMult32BitsPor13753ALa2 ;Cerrar ciclo.
        and eax,eax                    ;El DWord mas significativo vale cero?
        jnz short CantidadPalabrasDelTerminoOk32 ;Saltar si no es asi.
        inc PrimWord                   ;El primer Dword es el siguiente.
CantidadPalabrasDelTerminoOk32:
        add divisor32,2                ;El divisor 3,5,7,... de la formula
                                       ;debe ser incrementado.
        call PonerPotenciasBinarias32Bits ;Poner las potencias binarias en
                                       ;el termino.
        mov cx,CantWords               ;CX = Cantidad de DWords.
        lea si,[di+offset comienzo - 100h - (offset buffer[16384])]
        shr si,2
        sub cx,si
        jbe short BinarioADecimal32
        xor edx,edx                    ;Inicializar resto de la division.
        mov ebx,divisor32              ;Obtener divisor.
        cld                            ;Procesar en direcciones ascendentes
                                       ;(para realizar una division se deben
                                       ;tomar primero los DWords mas signif.)
        mov si,di
        add di,16384
        push cx
CicloDivisionPorValorVariable32:
        lodsd
        div ebx
        stosd
        loop CicloDivisionPorValorVariable32
        pop cx
        mov di,CantWords
        shl di,2
        lea si,buffer[di+32768-4]
        add di,offset buffer[-4]
        std
        ret
;----------------------------------------------------------------------------
BinarioADecimal32:
        pop ax                 ;No volver a rutina llamadora.
        mov ax,4400h           ;Llamada IOCTL.
        mov bx,1               ;Handle STDOUT.
        int 21h
        test dl,dl             ;Bit 7 de DL = 1: Dispositivo.
                               ;            = 0: Archivo.
        jns short TextoFormateado? ;Saltar si se redirecciono STDOUT.
        mov MostrarEnInt8,0    ;No mostrar texto en INT 8.
        mov dx,offset TiempoTotal
        mov cx,2
        mov bx,cx              ;BX = Handle STDERR.
        mov ah,40h
        int 21h
TextoFormateado?:
        test Indicadores,Formato       ;El texto debe estar formateado?
        jnz short ComenzarConversion   ;Saltar si es asi.
        mov dx,offset bufferdos
        mov cx,2                       ;Mostrar 3,
        mov bx,handle
        mov ah,40h
        int 21h
;Convertir de binario a decimal.
ComenzarConversion:
        cmp XMS_Instalado,0            ;Se esta utilizando memoria extendida?
        jz short MultiplicarEnRAMConv  ;Saltar si no es asi.
        mov eax,CantDWords             ;EAX = Cantidad de palabras del numero.
        shl eax,2                      ;Convertir a bytes.
        mov UltDWord,eax
MultiplicarXMSPor5ALa13Por2:
        mov ebx,UltDWord               ;EBX = Offset de la ultima DWord.
        mov eax,CantDWords             ;EAX = Cantidad de DWords a procesar.
        mov ebp,eax
        shr ebp,14 - 2                 ;BP = Cantidad de bloques completos.
        and ax,0FFFh
        movzx ecx,ax                   ;CX = Cantidad de DWords en el
                                       ;     ultimo bloque de 16 KB.
        shl cx,2                       ;Convertir a bytes.
        sub ebx,ecx                    ;Apuntar al principio del bloque.
        shr cx,2                       ;Convertir nuevamente a DWords.
        jnz short BloqueIncompleto
        mov cx,1000h                   ;CX = Cantidad de DWords en bloque.
        sub ebx,4000h                  ;Apuntar al bloque mas significativo.
        dec bp                         ;Indicar que hay un bloque menos.
BloqueIncompleto:
        mov OffsetXMSDestino,ebx
        xor ebx,ebx                    ;Inicializar arrastre de DWord menos
                                       ;significativo.
CicloMultEnConvBloqueXMS:
        mov esi,5*5*5*5*5*5*5*5*5*5*5*5*5*2  ;Multiplicador.
        push bp
        call SubrMultiplicarBloqueXMS  ;Realizar la multiplicacion en el
                                       ;ultimo bloque.
        pop bp                         ;Restaurar la cantidad de bloques
                                       ;completos.
        sub bp,1                       ;Hay algun bloque completo?
        mov cx,1000h                   ;Cantidad de DWords en el bloque.
        jnc short CicloMultEnConvBloqueXMS ;Saltar si no es asi.
        mov di,offset buffer[-4]
        jmp short FinMultEnConv
MultiplicarEnRAMConv:
        std
        mov ax,CantWords
        shl ax,2
        add ax,offset buffer[-4]
        mov UltWord,ax
MultiplicarPor5ALa13Por2:
        mov cx,CantWords
        mov di,UltWord
        mov esi,5*5*5*5*5*5*5*5*5*5*5*5*5*2
        xor ebx,ebx
CicloMultPor5ALa13Por2:
        mov eax,[di]
        mul esi
        add eax,ebx
        adc edx,0
        mov ebx,edx
        stosd
        loop CicloMultPor5ALa13Por2
FinMultEnConv:
        dec contador
        jns short DecrementarCantidadDePalabras32
        mov contador,31
        jmp short ObtenerSiguientes13Digitos
DecrementarCantidadDePalabras32:
        sub UltDWord,4
        dec CantDWords
ObtenerSiguientes13Digitos:
        mov cl,ContadorIzquierdo
        mov esi,[di+4]
        mov edx,[di+8]
        shld ebx,esi,cl
        shld esi,edx,cl
        shrd esi,ebx,20
        shr ebx,20
        mov eax,1 shl 20
        xor edx,edx
        shrd edx,eax,cl
        shr eax,cl
        sub edx,1
        sbb eax,0
        and [di+8],edx
        and [di+4],eax
        add cl,12
        cmp cl,32
        jb short seguir3
        sub cl,32
        dec CantDWords
seguir3:mov ContadorIzquierdo,cl
        mov di,offset bufferdos+12 ;DI = Puntero al final del buffer
                                ;temporario de digitos en pantalla.
        call EBX_ESI_ADecimal32Bits
        mov ecx,digitos32
        cmp ecx,13
        jbe short CantidadDeDigitosAMostrarOK32
        mov cx,13
CantidadDeDigitosAMostrarOK32:
        test Indicadores,Formato       ;Hay que formatear el texto?
        jnz short FormatearTexto32Bits ;Saltar si es asi.
        jcxz FinConversion32           ;Saltar si se indico cero decimales.
        mov bx,handle
        mov dx,offset bufferdos
        mov ah,40h
        int 21h
        sub digitos32,13
Otros13Digitos:
        jbe short FinConversion32
        cmp XMS_Instalado,0            ;Se esta utilizando memoria extendida?
        jz MultiplicarPor5ALa13Por2    ;Saltar si no es asi.
        mov ebx,OffsetXMSFuente
        mov OffsetFuenteRAMConv,offset buffer
        call SubrEnviarDatosHaciaXMS
        jmp MultiplicarXMSPor5ALa13Por2
FormatearTexto32Bits:
        jcxz UltimaLineaTextFormateado32Bits
        call DigitosFormateados32Bits
        cmp Digitos32,0
        jnz Otros13Digitos
UltimaLineaTextFormateado32Bits:
        call MostrarBufferEnPantalla32Bits
FinConversion32:
        cmp XMS_Instalado,0    ;Se uso XMS?
        jz short Final32       ;Saltar si no es asi.
        mov dx,XMS_Handle
        mov ah,0Ah             ;Liberar el handle de XMS.
        call XMS
Final32:
        mov dx,OldInt8Offset
        mov ds,OldInt8Segment
        mov ax,2508h
        int 21h
        push cs
        pop ds
        mov dx,offset TiempoTotal ;CR - LF
        mov cx,2
        mov bx,cx              ;Handle STDERR.
        mov ah,40h
        int 21h
        mov dx,offset TiempoTotal
        mov cx,LongitudTiempoTotal
        mov bx,2               ;Handle STDERR.
        mov ah,40h
        int 21h
        call obtencur
        call MostrarTiempo
        mov dx,offset TiempoTotal ;CR - LF
        mov cx,2
        mov bx,cx              ;Handle STDERR.
        mov ah,40h
        int 21h
        mov ax,4c00h           ;Fin del programa.
        int 21h
;----------------------------------------------------------------------------
; Manejador de la interrupcion 8 (timer). Debe mostrar la cantidad de digitos
; y el tiempo transcurrido.
;
Int8Handler:
        push ds                        ;Preservar los registros utilizados.
        push es
        push esi
        push di
        push bx
        push eax
        push edx
        push cx
        push cs                        ;Cargar el registro de segmento de
        pop ds                         ;datos.
        cld
        mov di,offset FraccSegundo
        add dword ptr [di],235903917   ;2^32 / 18,20648193
        jnc short CantDigitosAPantalla
        dec di                         ;Apuntar a los segundos.
        inc byte ptr [di]              ;Incrementar segundos.
        cmp byte ptr [di],60           ;Se termino el minuto?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
        dec di                         ;Apuntar a los minutos.
        inc byte ptr [di]              ;Incrementar minutos.
        cmp byte ptr [di],60           ;Se termino la hora?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
        dec di                         ;Apuntar a las horas.
        inc byte ptr [di]              ;Incrementar horas.
        cmp byte ptr [di],24           ;Se termino el dia?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
        dec di                         ;Apuntar a dias.
        inc byte ptr [di]              ;Incrementar horas.
        cmp byte ptr [di],100          ;Pasaron cien dias?
        jnz short TiempoActualizado    ;Saltar si no es asi.
        mov byte ptr [di],0
TiempoActualizado:
        cmp MostrarEnInt8,0            ;Hay que mostrar numero y tiempo?
        jz short FinInter8             ;Salir de la interrupcion si no es asi.
        les di,CursorVideo             ;Obtener la posicion del cursor.
        add di,15*2                    ;Apuntar al tiempo.
        call MostrarTiempo             ;Mostrar tiempo transcurrido.
CantDigitosAPantalla:
        cmp MostrarEnInt8,0            ;Hay que mostrar numero y tiempo?
        jz short FinInter8             ;Salir de la interrupcion si no es asi.
        les di,CursorVideo             ;Obtener la posicion del cursor.
        mov eax,3878741959             ;2^32 * log 8.
        mul Divisor32                  ;EDX = Cantidad de digitos.
        mov bl,0
        mov esi,1000000000
        call Div1Digito
        mov esi,100000000
        call Div1Digito
        mov esi,10000000
        call Div1Digito
        mov esi,1000000
        call Div1Digito
        mov esi,100000
        call Div1Digito
        mov esi,10000
        call Div1Digito
        mov si,1000
        call Div1Digito
        mov si,100
        call Div1Digito
        mov si,10
        call Div1Digito
        add dl,"0"
        mov es:[di],dl
FinInter8:
        pop cx                  ;Restaurar los registros afectados.
        pop edx
        pop eax
        pop bx
        pop di
        pop esi
        pop es
        pop ds
        jmp cs:OldInt8Handler   ;Ir al viejo manejador de INT 8.
;---------------------------------------------------------------------------
; Subrutina para obtener la posicin del cursor en la pantalla.
; Salida: ES:DI = Posicion en buffer de video que corresponde al cursor.
;
obtencur:mov dx,port6845     ;DX = Port de control del controlador de video.
        mov al,0Eh           ;Indicarle que hay que leer la parte alta
        out dx,al            ;(MSB) de la posicin del cursor.
        jmp $+2
        jmp $+2
        inc dx               ;Apuntar al port de datos del 6845.
        in al,dx             ;Leer el valor.
        mov ah,al            ;Ponerlo en la parte alta de AX.
        jmp $+2
        jmp $+2
        dec dx               ;Apuntar al port de control del 6845,
        mov al,0Fh           ;Indicarle que hay que leer la parte baja
        out dx,al            ;(LSB) de la posicin del cursor.
        jmp $+2
        jmp $+2
        inc dx               ;Apuntar al port de datos del 6845.
        in al,dx             ;Leer el valor.
        add ax,ax
        mov di,ax            ;Poner el resultado en SI.
        mov ax,0B800h
        mov es,ax
        ret                  ;Fin de la subrutina.
;----------------------------------------------------------------------------
; Subrutina MostrarTiempo: Mostrar el tiempo transcurrido.
; Entrada: ES:DI = Posicion en buffer de video.
;
MostrarTiempo:
        mov si,offset Dias
        mov cx,4
        cld
CicloMostrarTiempo:
        lodsb
        aam
        add ax,"00"
        mov es:[di],ah
        mov es:[di+2],al
        mov al,[si-5]
        mov es:[di+4],al
        mov byte ptr es:[di+6]," "
        add di,8
        loop CicloMostrarTiempo
        ret
;-----------------------------------------------------------------------------
Div1Digito:
        mov eax,edx
        xor edx,edx
        div esi
        add al,"0"
        cmp al,"0"
        jnz short MostrarDigito
        test bl,bl
        jz short ApuntarSiguienteDigito
MostrarDigito:
        mov bl,1
        stosb
        inc di
ApuntarSiguienteDigito:
        ret
;-----------------------------------------------------------------------------
; Subrutina HabilitarInterrupcion: Instala la nueva interrupcion 8.
;
HabilitarInterrupcion:
        push ds
        push es
        mov dx,offset TextoSegundaLinea
        mov cx,LongitudTextoSegundaLinea
        mov bx,2                ;Handle STDERR.
        mov ah,40h
        int 21h
        mov ax,3508h
        int 21h                 ;ES:BX = Manejador del viejo INT 8.
        mov OldInt8Offset,bx
        mov OldInt8Segment,es
        mov ax,2508h
        push cs
        pop ds
        mov dx,offset Int8Handler
        int 21h
        call obtencur
        mov CursorVideoOffset,di
        mov CursorVideoSegment,es
        pop es
        pop ds
        ret
;-----------------------------------------------------------------------------
; Subrutina DigitosFormateados32Bits: Halla un digito del numero y, si esta
; habilitado con el switch /F, realiza el formateo de texto.
;
; Entrada: bufferdos: 13 digitos en ASCII.
;          CX = Cantidad de digitos a mostrar.
;          DI = Puntero al buffer donde se almacenan los digitos en ASCII.
;
DigitosFormateados32Bits:
        cld
        mov si,offset bufferdos
        mov ah,0         ;Mostrar los ceros mas significativos como espacios.
        test Indicadores,MostrarCantDigitos
        jnz short CicloMostrarSinFormatear
        test Indicadores,Formato
        jnz short MostrarFormateado32Bits
        mov ah,1         ;Mostrar los ceros mas significativos como ceros.
CicloMostrarSinFormatear:
        lodsb
        cmp al,"0"
        jnz short DigitoSignificativo32Bits
        test ah,ah
        jnz short DigitoSignificativo32Bits
        cmp si,offset bufferdos[13]
        jz short MostrarDigito32Bits
        mov al," "
        jmp short MostrarDigito32Bits
DigitoSignificativo32Bits:
        mov ah,1
MostrarDigito32Bits:
        stosb
        loop CicloMostrarSinFormatear
        std
        ret
MostrarFormateado32Bits:
        dec digitos32
        mov di,PosicionDigitoEnBuffer
        mov bx,InformacionDeFormateado  ;BH = Numero de grupo (0-9).
                                        ;BL = Digito dentro del grupo (1-5).
        inc bx                          ;Indicar siguiente digito del grupo.
        cmp bl,6                        ;Fin del grupo?
        jnz short PonerDigitoFormateado32Bits  ;Saltar si no es asi.
        mov bl,1                        ;Indicar inicio del grupo.
        inc bh                          ;Indicar siguiente grupo.
        cmp bh,10                       ;Fin de la linea?
        jnz short PonerEspacio32Bits    ;Ir a mostrar espacio si no es asi.
        push si
        call MostrarBufferEnPantalla32Bits
        pop si
        mov di,offset BufferNumeroFormateado 
                                     ;Borrar el buffer de digitos formateados.
        push cx
        cld
        mov cx,75
        mov al," "
        rep stosb
        pop cx
        mov bx,1
        mov di,offset BufferNumeroFormateado + 1
                                       ;Apuntar al segundo byte del buffer.
PonerEspacio32Bits:
        inc di          ;Apuntar al siguiente caracter.
PonerDigitoFormateado32Bits:
        movsb           ;Copiar el digito en buffer de digitos formateados.
        mov PosicionDigitoEnBuffer,di
        mov InformacionDeFormateado,bx
        inc DigitosMostrados32Bits ;Indicar que se mostro un digito mas.
        loop MostrarFormateado32Bits
        std
        ret                     ;Fin de la subrutina.
;---------------------------------------------------------------------------
MostrarBufferEnPantalla32Bits:
        push cx
        or Indicadores,MostrarCantDigitos
        push esi        
        mov esi,DigitosMostrados32Bits
        xor ebx,ebx
        mov di,offset BufferNumeroFormateado[74]
                                ;DI = Puntero al final del buffer
                                ;     del numero a mostrar a la derecha.
        call EBX_ESI_ADecimal32Bits
        mov cx,12
CicloEliminarCerosIzq:
        inc di
        cmp byte ptr [di],"0"
        jnz short CerosEliminados
        mov byte ptr [di]," "
        loop CicloEliminarCerosIzq
CerosEliminados:
        pop esi
        and Indicadores,not MostrarCantDigitos
        mov bx,handle
        mov dx,offset BufferNumeroFormateado
        mov cx,79
        mov ah,40h
        int 21h
        pop cx
        ret
;------------------------------------------------------------------------------
; Conversion binario (EBX:ESI = 64 bits) a decimal (hasta 13 digitos).
;
; Entrada: EBX = Doble palabra mas significativa.
;          EDI = Doble palabra menos significativa.
; Salida:  bufferdos: Numero en ASCII.
;
EBX_ESI_ADecimal32Bits:
        mov ebp,10              ;EBP = Divisor.
        mov cx,13               ;Cantidad de digitos del numero.
bucle_bin_a_dec:
        xor edx,edx             ;Extender el dividendo.
        mov eax,ebx             ;Cargar la parte alta del dividendo.
        div ebp                 ;Realizar la primera parte de la division.
        mov ebx,eax             ;Salvar la nueva parte alta del dividendo.
        mov eax,esi             ;Cargar la parte baja del dividendo.
        div ebp                 ;Realizar la segunda parte de la division.
        mov esi,eax             ;Salvar la nueva parte baja del dividendo.
        add dl,"0"              ;Convertir a ASCII.
        mov [di],dl             ;Guardar el resto en el buffer temporario
                                ;(de esta manera se obtiene el digito
                                ;"de la derecha" o menos significativo).
        dec di                  ;Apuntar al digito anterior.
        loop bucle_bin_a_dec    ;Saltar si no se acabo el buffer.
        ret                     ;Fin de la subrutina.
;---------------------------------------------------------------------------
        align 4
buffer  label word
codigo  ends
        end comienzo
